# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from io import StringIO
import logging
from os.path import join, normpath
import pickle
import os

from pyomo.common.dependencies import dill_available
from pyomo.common.fileutils import import_file, PYOMO_ROOT_DIR
from pyomo.common.log import LoggingIntercept
import pyomo.common.unittest as unittest
from pyomo.core.expr.compare import (
    assertExpressionsEqual,
    assertExpressionsStructurallyEqual,
)

from pyomo.environ import (
    BooleanVar,
    ConcreteModel,
    Constraint,
    LogicalConstraint,
    NonNegativeIntegers,
    SolverFactory,
    Suffix,
    TransformationFactory,
    value,
    Var,
    Block,
)
from pyomo.gdp import Disjunct, Disjunction, GDP_Error
from pyomo.gdp.tests.common_tests import (
    check_linear_coef,
    check_nested_disjuncts_in_flat_gdp,
    check_obj_in_active_tree,
    check_pprint_equal,
)
from pyomo.gdp.tests.models import make_indexed_equality_model
from pyomo.repn import generate_standard_repn

gurobi_available = (
    SolverFactory('gurobi').available(exception_flag=False)
    and SolverFactory('gurobi').license_is_valid()
)
gurobi_direct_available = (
    SolverFactory('gurobi_direct_v2').available(exception_flag=False)
    and SolverFactory('gurobi_direct_v2').license_is_valid()
)
exdir = normpath(join(PYOMO_ROOT_DIR, 'examples', 'gdp'))


class CommonTests(unittest.TestCase):
    def check_pretty_bound_constraints(self, cons, var, bounds, lb):
        self.assertEqual(value(cons.upper), 0)
        self.assertIsNone(cons.lower)
        repn = generate_standard_repn(cons.body)
        self.assertTrue(repn.is_linear())
        self.assertEqual(len(repn.linear_vars), len(bounds) + 1)
        self.assertEqual(repn.constant, 0)
        if lb:
            check_linear_coef(self, repn, var, -1)
            for disj, bnd in bounds.items():
                check_linear_coef(self, repn, disj.binary_indicator_var, bnd)
        else:
            check_linear_coef(self, repn, var, 1)
            for disj, bnd in bounds.items():
                check_linear_coef(self, repn, disj.binary_indicator_var, -bnd)


class LinearModelDecisionTreeExample(CommonTests):
    def make_model(self):
        m = ConcreteModel()
        m.x1 = Var(bounds=(-10, 10))
        m.x2 = Var(bounds=(-20, 20))
        m.d = Var(bounds=(-1000, 1000))

        m.d1 = Disjunct()
        m.d1.x1_bounds = Constraint(expr=(0.5, m.x1, 2))
        m.d1.x2_bounds = Constraint(expr=(0.75, m.x2, 3))
        m.d1.func = Constraint(expr=m.x1 + m.x2 == m.d)

        m.d2 = Disjunct()
        m.d2.x1_bounds = Constraint(expr=(0.65, m.x1, 3))
        m.d2.x2_bounds = Constraint(expr=(3, m.x2, 10))
        m.d2.func = Constraint(expr=2 * m.x1 + 4 * m.x2 + 7 == m.d)

        m.d3 = Disjunct()
        m.d3.x1_bounds = Constraint(expr=(2, m.x1, 10))
        m.d3.x2_bounds = Constraint(expr=(0.55, m.x2, 1))
        m.d3.func = Constraint(expr=m.x1 - 5 * m.x2 - 3 == m.d)

        m.disjunction = Disjunction(expr=[m.d1, m.d2, m.d3])

        return m

    def get_Ms(self, m):
        return {
            (m.d1.x1_bounds, m.d2): (0.15, 1),
            (m.d1.x2_bounds, m.d2): (2.25, 7),
            (m.d1.x1_bounds, m.d3): (1.5, 8),
            (m.d1.x2_bounds, m.d3): (-0.2, -2),
            (m.d2.x1_bounds, m.d1): (-0.15, -1),
            (m.d2.x2_bounds, m.d1): (-2.25, -7),
            (m.d2.x1_bounds, m.d3): (1.35, 7),
            (m.d2.x2_bounds, m.d3): (-2.45, -9),
            (m.d3.x1_bounds, m.d1): (-1.5, -8),
            (m.d3.x2_bounds, m.d1): (0.2, 2),
            (m.d3.x1_bounds, m.d2): (-1.35, -7),
            (m.d3.x2_bounds, m.d2): (2.45, 9),
            (m.d1.func, m.d2): (-40, -16.65),
            (m.d1.func, m.d3): (6.3, 9),
            (m.d2.func, m.d1): (9.75, 18),
            (m.d2.func, m.d3): (16.95, 29),
            (m.d3.func, m.d1): (-21, -7.5),
            (m.d3.func, m.d2): (-103, -37.65),
        }

    def check_untightened_bounds_constraint(
        self, cons, var, parent_disj, disjunction, Ms, lower=None, upper=None
    ):
        repn = generate_standard_repn(cons.body)
        self.assertTrue(repn.is_linear())
        self.assertEqual(len(repn.linear_vars), 3)
        self.assertIsNone(cons.lower)
        self.assertEqual(value(cons.upper), 0)
        if lower is not None:
            self.assertEqual(repn.constant, lower)
            check_linear_coef(self, repn, var, -1)
            for disj in disjunction.disjuncts:
                if disj is not parent_disj:
                    check_linear_coef(
                        self, repn, disj.binary_indicator_var, Ms[disj] - lower
                    )
        if upper is not None:
            self.assertEqual(repn.constant, -upper)
            check_linear_coef(self, repn, var, 1)
            for disj in disjunction.disjuncts:
                if disj is not parent_disj:
                    check_linear_coef(
                        self, repn, disj.binary_indicator_var, -Ms[disj] + upper
                    )

    def check_all_untightened_bounds_constraints(self, m, mbm):
        # d1.x1_bounds
        cons = mbm.get_transformed_constraints(m.d1.x1_bounds)
        self.assertEqual(len(cons), 2)
        lower = cons[0]
        check_obj_in_active_tree(self, lower)
        self.check_untightened_bounds_constraint(
            lower, m.x1, m.d1, m.disjunction, {m.d2: 0.65, m.d3: 2}, lower=0.5
        )
        upper = cons[1]
        check_obj_in_active_tree(self, upper)
        self.check_untightened_bounds_constraint(
            upper, m.x1, m.d1, m.disjunction, {m.d2: 3, m.d3: 10}, upper=2
        )

        # d1.x2_bounds
        cons = mbm.get_transformed_constraints(m.d1.x2_bounds)
        self.assertEqual(len(cons), 2)
        lower = cons[0]
        check_obj_in_active_tree(self, lower)
        self.check_untightened_bounds_constraint(
            lower, m.x2, m.d1, m.disjunction, {m.d2: 3, m.d3: 0.55}, lower=0.75
        )
        upper = cons[1]
        check_obj_in_active_tree(self, upper)
        self.check_untightened_bounds_constraint(
            upper, m.x2, m.d1, m.disjunction, {m.d2: 10, m.d3: 1}, upper=3
        )

        # d2.x1_bounds
        cons = mbm.get_transformed_constraints(m.d2.x1_bounds)
        self.assertEqual(len(cons), 2)
        lower = cons[0]
        check_obj_in_active_tree(self, lower)
        self.check_untightened_bounds_constraint(
            lower, m.x1, m.d2, m.disjunction, {m.d1: 0.5, m.d3: 2}, lower=0.65
        )
        upper = cons[1]
        check_obj_in_active_tree(self, upper)
        self.check_untightened_bounds_constraint(
            upper, m.x1, m.d2, m.disjunction, {m.d1: 2, m.d3: 10}, upper=3
        )

        # d2.x2_bounds
        cons = mbm.get_transformed_constraints(m.d2.x2_bounds)
        self.assertEqual(len(cons), 2)
        lower = cons[0]
        check_obj_in_active_tree(self, lower)
        self.check_untightened_bounds_constraint(
            lower, m.x2, m.d2, m.disjunction, {m.d1: 0.75, m.d3: 0.55}, lower=3
        )
        upper = cons[1]
        self.check_untightened_bounds_constraint(
            upper, m.x2, m.d2, m.disjunction, {m.d1: 3, m.d3: 1}, upper=10
        )

        # d3.x1_bounds
        cons = mbm.get_transformed_constraints(m.d3.x1_bounds)
        self.assertEqual(len(cons), 2)
        lower = cons[0]
        self.check_untightened_bounds_constraint(
            lower, m.x1, m.d3, m.disjunction, {m.d1: 0.5, m.d2: 0.65}, lower=2
        )
        upper = cons[1]
        check_obj_in_active_tree(self, upper)
        self.check_untightened_bounds_constraint(
            upper, m.x1, m.d3, m.disjunction, {m.d1: 2, m.d2: 3}, upper=10
        )

        # d3.x2_bounds
        cons = mbm.get_transformed_constraints(m.d3.x2_bounds)
        self.assertEqual(len(cons), 2)
        lower = cons[0]
        check_obj_in_active_tree(self, lower)
        self.check_untightened_bounds_constraint(
            lower, m.x2, m.d3, m.disjunction, {m.d1: 0.75, m.d2: 3}, lower=0.55
        )
        upper = cons[1]
        check_obj_in_active_tree(self, upper)
        self.check_untightened_bounds_constraint(
            upper, m.x2, m.d3, m.disjunction, {m.d1: 3, m.d2: 10}, upper=1
        )

    def check_linear_func_constraints(self, m, mbm, Ms=None):
        if Ms is None:
            Ms = self.get_Ms(m)

        # d1.func
        cons = mbm.get_transformed_constraints(m.d1.func)
        self.assertEqual(len(cons), 2)
        lower = cons[0]
        check_obj_in_active_tree(self, lower)
        self.assertEqual(value(lower.upper), 0)
        self.assertIsNone(lower.lower)
        repn = generate_standard_repn(lower.body)
        self.assertTrue(repn.is_linear())
        self.assertEqual(len(repn.linear_vars), 5)
        self.assertEqual(repn.constant, 0)
        check_linear_coef(self, repn, m.x1, -1)
        check_linear_coef(self, repn, m.x2, -1)
        check_linear_coef(self, repn, m.d, 1)
        check_linear_coef(self, repn, m.d2.binary_indicator_var, Ms[m.d1.func, m.d2][0])
        check_linear_coef(self, repn, m.d3.binary_indicator_var, Ms[m.d1.func, m.d3][0])
        upper = cons[1]
        check_obj_in_active_tree(self, upper)
        self.assertEqual(value(upper.upper), 0)
        self.assertIsNone(upper.lower)
        repn = generate_standard_repn(upper.body)
        self.assertTrue(repn.is_linear())
        self.assertEqual(len(repn.linear_vars), 5)
        self.assertEqual(repn.constant, 0)
        check_linear_coef(self, repn, m.x1, 1)
        check_linear_coef(self, repn, m.x2, 1)
        check_linear_coef(self, repn, m.d, -1)
        check_linear_coef(
            self, repn, m.d2.binary_indicator_var, -Ms[m.d1.func, m.d2][1]
        )
        check_linear_coef(
            self, repn, m.d3.binary_indicator_var, -Ms[m.d1.func, m.d3][1]
        )

        # d2.func
        cons = mbm.get_transformed_constraints(m.d2.func)
        self.assertEqual(len(cons), 2)
        lower = cons[0]
        check_obj_in_active_tree(self, lower)
        self.assertEqual(value(lower.upper), 0)
        self.assertIsNone(lower.lower)
        repn = generate_standard_repn(lower.body)
        self.assertTrue(repn.is_linear())
        self.assertEqual(len(repn.linear_vars), 5)
        self.assertEqual(repn.constant, -7)
        check_linear_coef(self, repn, m.x1, -2)
        check_linear_coef(self, repn, m.x2, -4)
        check_linear_coef(self, repn, m.d, 1)
        check_linear_coef(self, repn, m.d1.binary_indicator_var, Ms[m.d2.func, m.d1][0])
        check_linear_coef(self, repn, m.d3.binary_indicator_var, Ms[m.d2.func, m.d3][0])
        upper = cons[1]
        check_obj_in_active_tree(self, upper)
        self.assertEqual(value(upper.upper), 0)
        self.assertIsNone(upper.lower)
        repn = generate_standard_repn(upper.body)
        self.assertTrue(repn.is_linear())
        self.assertEqual(len(repn.linear_vars), 5)
        self.assertEqual(repn.constant, 7)
        check_linear_coef(self, repn, m.x1, 2)
        check_linear_coef(self, repn, m.x2, 4)
        check_linear_coef(self, repn, m.d, -1)
        check_linear_coef(
            self, repn, m.d1.binary_indicator_var, -Ms[m.d2.func, m.d1][1]
        )
        check_linear_coef(
            self, repn, m.d3.binary_indicator_var, -Ms[m.d2.func, m.d3][1]
        )

        # d3.func
        cons = mbm.get_transformed_constraints(m.d3.func)
        self.assertEqual(len(cons), 2)
        lower = cons[0]
        check_obj_in_active_tree(self, lower)
        self.assertEqual(value(lower.upper), 0)
        self.assertIsNone(lower.lower)
        repn = generate_standard_repn(lower.body)
        self.assertTrue(repn.is_linear())
        self.assertEqual(len(repn.linear_vars), 5)
        self.assertEqual(repn.constant, 3)
        check_linear_coef(self, repn, m.x1, -1)
        check_linear_coef(self, repn, m.x2, 5)
        check_linear_coef(self, repn, m.d, 1)
        check_linear_coef(self, repn, m.d1.binary_indicator_var, Ms[m.d3.func, m.d1][0])
        check_linear_coef(self, repn, m.d2.binary_indicator_var, Ms[m.d3.func, m.d2][0])
        upper = cons[1]
        check_obj_in_active_tree(self, upper)
        self.assertEqual(value(upper.upper), 0)
        self.assertIsNone(upper.lower)
        repn = generate_standard_repn(upper.body)
        self.assertTrue(repn.is_linear())
        self.assertEqual(len(repn.linear_vars), 5)
        self.assertEqual(repn.constant, -3)
        check_linear_coef(self, repn, m.x1, 1)
        check_linear_coef(self, repn, m.x2, -5)
        check_linear_coef(self, repn, m.d, -1)
        check_linear_coef(
            self, repn, m.d1.binary_indicator_var, -Ms[m.d3.func, m.d1][1]
        )
        check_linear_coef(
            self, repn, m.d2.binary_indicator_var, -Ms[m.d3.func, m.d2][1]
        )

    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    def test_calculated_Ms_correct(self):
        # Calculating all the Ms is expensive, so we just do it in this one test
        # and then specify them for most of the others
        m = self.make_model()
        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m, reduce_bound_constraints=False)

        self.check_all_untightened_bounds_constraints(m, mbm)
        self.check_linear_func_constraints(m, mbm)

        self.assertStructuredAlmostEqual(mbm.get_all_M_values(m), self.get_Ms(m))

    def test_transformed_constraints_correct_Ms_specified(self):
        m = self.make_model()
        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m, bigM=self.get_Ms(m), reduce_bound_constraints=False)

        self.check_all_untightened_bounds_constraints(m, mbm)
        self.check_linear_func_constraints(m, mbm)

    def test_local_var_suffix_ignored(self):
        m = self.make_model()
        m.y = Var(bounds=(2, 5))
        m.d1.another_thing = Constraint(expr=m.y == 3)
        m.d1.LocalVars = Suffix(direction=Suffix.LOCAL)
        m.d1.LocalVars[m.d1] = m.y

        mbigm = TransformationFactory('gdp.mbigm')
        mbigm.apply_to(
            m, reduce_bound_constraints=True, only_mbigm_bound_constraints=True
        )

        cons = mbigm.get_transformed_constraints(m.d1.x1_bounds)
        self.check_pretty_bound_constraints(
            cons[0], m.x1, {m.d1: 0.5, m.d2: 0.65, m.d3: 2}, lb=True
        )
        self.check_pretty_bound_constraints(
            cons[1], m.x1, {m.d1: 2, m.d2: 3, m.d3: 10}, lb=False
        )

        cons = mbigm.get_transformed_constraints(m.d1.x2_bounds)
        self.check_pretty_bound_constraints(
            cons[0], m.x2, {m.d1: 0.75, m.d2: 3, m.d3: 0.55}, lb=True
        )
        self.check_pretty_bound_constraints(
            cons[1], m.x2, {m.d1: 3, m.d2: 10, m.d3: 1}, lb=False
        )

        cons = mbigm.get_transformed_constraints(m.d1.another_thing)
        self.assertEqual(len(cons), 2)
        self.check_pretty_bound_constraints(
            cons[0], m.y, {m.d1: 3, m.d2: 2, m.d3: 2}, lb=True
        )
        self.check_pretty_bound_constraints(
            cons[1], m.y, {m.d1: 3, m.d2: 5, m.d3: 5}, lb=False
        )

    def test_pickle_transformed_model(self):
        m = self.make_model()
        TransformationFactory('gdp.mbigm').apply_to(m, bigM=self.get_Ms(m))

        # pickle and unpickle the transformed model
        unpickle = pickle.loads(pickle.dumps(m))

        check_pprint_equal(self, m, unpickle)

    def test_mappings_between_original_and_transformed_components(self):
        m = self.make_model()
        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m, bigM=self.get_Ms(m), reduce_bound_constraints=False)

        d1_block = m.d1.transformation_block
        self.assertIs(mbm.get_src_disjunct(d1_block), m.d1)
        d2_block = m.d2.transformation_block
        self.assertIs(mbm.get_src_disjunct(d2_block), m.d2)
        d3_block = m.d3.transformation_block
        self.assertIs(mbm.get_src_disjunct(d3_block), m.d3)

        for disj in [m.d1, m.d2, m.d3]:
            for comp in ['x1_bounds', 'x2_bounds', 'func']:
                original_cons = disj.component(comp)
                transformed = mbm.get_transformed_constraints(original_cons)
                for cons in transformed:
                    self.assertIn(original_cons, mbm.get_src_constraints(cons))

    def test_algebraic_constraints(self):
        m = self.make_model()
        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m, bigM=self.get_Ms(m), reduce_bound_constraints=False)

        self.assertIsNotNone(m.disjunction.algebraic_constraint)
        xor = m.disjunction.algebraic_constraint
        self.assertIs(mbm.get_src_disjunction(xor), m.disjunction)

        self.assertEqual(value(xor.lower), 1)
        self.assertEqual(value(xor.upper), 1)
        repn = generate_standard_repn(xor.body)
        self.assertTrue(repn.is_linear())
        self.assertEqual(value(repn.constant), 0)
        self.assertEqual(len(repn.linear_vars), 3)
        check_linear_coef(self, repn, m.d1.binary_indicator_var, 1)
        check_linear_coef(self, repn, m.d2.binary_indicator_var, 1)
        check_linear_coef(self, repn, m.d3.binary_indicator_var, 1)
        check_obj_in_active_tree(self, xor)

    def test_bounds_constraints_correct(self):
        m = self.make_model()

        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m, bigM=self.get_Ms(m), reduce_bound_constraints=True)

        # Check that all the constraints are mapped to the same transformed
        # constraints.
        cons = mbm.get_transformed_constraints(m.d1.x1_bounds)
        self.assertEqual(len(cons), 2)
        same = mbm.get_transformed_constraints(m.d2.x1_bounds)
        self.assertEqual(len(same), 2)
        self.assertIs(same[0], cons[0])
        self.assertIs(same[1], cons[1])
        sameagain = mbm.get_transformed_constraints(m.d3.x1_bounds)
        self.assertEqual(len(sameagain), 2)
        self.assertIs(sameagain[0], cons[0])
        self.assertIs(sameagain[1], cons[1])

        self.check_pretty_bound_constraints(
            cons[0], m.x1, {m.d1: 0.5, m.d2: 0.65, m.d3: 2}, lb=True
        )
        self.check_pretty_bound_constraints(
            cons[1], m.x1, {m.d1: 2, m.d2: 3, m.d3: 10}, lb=False
        )

        cons = mbm.get_transformed_constraints(m.d1.x2_bounds)
        self.assertEqual(len(cons), 2)
        same = mbm.get_transformed_constraints(m.d2.x2_bounds)
        self.assertEqual(len(same), 2)
        self.assertIs(same[0], cons[0])
        self.assertIs(same[1], cons[1])
        sameagain = mbm.get_transformed_constraints(m.d3.x2_bounds)
        self.assertEqual(len(sameagain), 2)
        self.assertIs(sameagain[0], cons[0])
        self.assertIs(sameagain[1], cons[1])

        self.check_pretty_bound_constraints(
            cons[0], m.x2, {m.d1: 0.75, m.d2: 3, m.d3: 0.55}, lb=True
        )
        self.check_pretty_bound_constraints(
            cons[1], m.x2, {m.d1: 3, m.d2: 10, m.d3: 1}, lb=False
        )

    def test_bound_constraints_correct_with_redundant_constraints(self):
        m = self.make_model()

        m.d1.bogus_x1_bounds = Constraint(expr=(0.25, m.x1, 2.25))
        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m, reduce_bound_constraints=True, bigM=self.get_Ms(m))

        cons = mbm.get_transformed_constraints(m.d1.x1_bounds)
        self.assertEqual(len(cons), 2)
        self.check_pretty_bound_constraints(
            cons[0], m.x1, {m.d1: 0.5, m.d2: 0.65, m.d3: 2}, lb=True
        )
        self.check_pretty_bound_constraints(
            cons[1], m.x1, {m.d1: 2, m.d2: 3, m.d3: 10}, lb=False
        )

        cons = mbm.get_transformed_constraints(m.d1.x2_bounds)
        self.assertEqual(len(cons), 2)
        self.check_pretty_bound_constraints(
            cons[0], m.x2, {m.d1: 0.75, m.d2: 3, m.d3: 0.55}, lb=True
        )
        self.check_pretty_bound_constraints(
            cons[1], m.x2, {m.d1: 3, m.d2: 10, m.d3: 1}, lb=False
        )

    def test_Ms_specified_as_args_honored(self):
        m = self.make_model()

        Ms = self.get_Ms(m)
        # now modify some of them--these might not be valid and I don't care
        Ms[m.d2.x2_bounds, m.d3] = (-100, 100)
        Ms[m.d3.func, m.d1] = [10, 20]

        mbigm = TransformationFactory('gdp.mbigm')
        mbigm.apply_to(m, bigM=Ms, reduce_bound_constraints=False)

        self.assertStructuredAlmostEqual(mbigm.get_all_M_values(m), Ms)
        self.check_linear_func_constraints(m, mbigm, Ms)

        # Just check the constraint we should have changed
        cons = mbigm.get_transformed_constraints(m.d2.x2_bounds)
        self.assertEqual(len(cons), 2)
        # This is a little backwards because I structured these so we give the
        # bound not the value of M. So the logic here is that if we want M to
        # turn out to be -100, we need b - 3 = -100, so we pretend the bound was
        # b=-97. The same logic holds for the next one too.
        self.check_untightened_bounds_constraint(
            cons[0], m.x2, m.d2, m.disjunction, {m.d1: 0.75, m.d3: -97}, lower=3
        )
        self.check_untightened_bounds_constraint(
            cons[1], m.x2, m.d2, m.disjunction, {m.d1: 3, m.d3: 110}, upper=10
        )

    # TODO: If Suffixes allow tuple keys then we can support them and it will
    # look something like this:
    # def test_Ms_specified_as_suffixes_honored(self):
    #     m = self.make_model()
    #     m.BigM = Suffix(direction=Suffix.LOCAL)
    #     m.BigM[(m.d2.x2_bounds, m.d3)] = (-100, 100)
    #     m.d3.BigM = Suffix(direction=Suffix.LOCAL)
    #     m.d3.BigM[(m.d3.func, m.d1)] = [10, 20]

    #     arg_Ms = self.get_Ms(m)
    #     # delete the keys we replaced above
    #     del arg_Ms[m.d2.x2_bounds, m.d3]
    #     del arg_Ms[m.d3.func, m.d1]

    #     mbigm = TransformationFactory('gdp.mbigm')
    #     mbigm.apply_to(m, bigM=arg_Ms)

    #     Ms = self.get_Ms(m)
    #     self.assertStructuredAlmostEqual(mbigm.get_all_M_values(m), Ms)
    #     self.check_linear_func_constraints(m, mbigm, Ms)

    #     # Just check the constraint we should have changed
    #     cons = mbigm.get_transformed_constraints(m.d2.x2_bounds)
    #     self.assertEqual(len(cons), 2)
    #     # This is a little backwards because I structured these so we give the
    #     # bound not the value of M. So the logic here is that if we want M to
    #     # turn out to be -100, we need b - 3 = -100, so we pretend the bound was
    #     # b=-97. The same logic holds for the next one too.
    #     self.check_untightened_bounds_constraint(cons[0], m.x2, m.d2,
    #                                              m.disjunction, {m.d1: 0.75,
    #                                                              m.d3: -97},
    #                                              lower=3)
    #     self.check_untightened_bounds_constraint(cons[1], m.x2, m.d2,
    #                                              m.disjunction, {m.d1: 3,
    #                                                              m.d3: 110},
    #                                              upper=10)

    def add_fourth_disjunct(self, m):
        m.disjunction.deactivate()

        # Add a disjunct
        m.d4 = Disjunct()
        m.d4.x1_ub = Constraint(expr=m.x1 <= 8)
        m.d4.x2_lb = Constraint(expr=m.x2 >= -5)

        # Make a four-term disjunction
        m.disjunction2 = Disjunction(expr=[m.d1, m.d2, m.d3, m.d4])

    def test_deactivated_disjunct(self):
        m = self.make_model()
        # Add a new thing and deactivate it
        self.add_fourth_disjunct(m)
        m.d4.deactivate()

        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m, bigM=self.get_Ms(m), reduce_bound_constraints=False)

        # we don't transform d1
        self.assertIsNone(m.d4.transformation_block)
        # and everything else is the same
        self.check_linear_func_constraints(m, mbm)
        self.check_all_untightened_bounds_constraints(m, mbm)

    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    def test_var_bounds_substituted_for_missing_bound_constraints(self):
        m = self.make_model()
        # Add a new thing with constraints that don't give both bounds on x1 and
        # x2
        self.add_fourth_disjunct(m)

        mbm = TransformationFactory('gdp.mbigm')
        # We will ignore the specified M values for the bounds constraints, but
        # issue a warning about what was unnecessary.
        out = StringIO()
        with LoggingIntercept(out, 'pyomo.gdp.mbigm'):
            mbm.apply_to(m, bigM=self.get_Ms(m), reduce_bound_constraints=True)

        warnings = out.getvalue()
        self.assertIn(
            "Unused arguments in the bigM map! "
            "These arguments were not used by the "
            "transformation:",
            warnings,
        )
        for cons, disj in [
            (m.d1.x1_bounds, m.d2),
            (m.d1.x2_bounds, m.d2),
            (m.d1.x1_bounds, m.d3),
            (m.d1.x2_bounds, m.d3),
            (m.d2.x1_bounds, m.d1),
            (m.d2.x2_bounds, m.d1),
            (m.d2.x1_bounds, m.d3),
            (m.d2.x2_bounds, m.d3),
            (m.d3.x1_bounds, m.d1),
            (m.d3.x2_bounds, m.d1),
            (m.d3.x1_bounds, m.d2),
            (m.d3.x2_bounds, m.d2),
        ]:
            self.assertIn("(%s, %s)" % (cons.name, disj.name), warnings)

        # check that the bounds constraints are right
        # for x1:
        cons = mbm.get_transformed_constraints(m.d1.x1_bounds)
        self.assertEqual(len(cons), 2)
        sameish = mbm.get_transformed_constraints(m.d4.x1_ub)
        self.assertEqual(len(sameish), 1)
        self.assertIs(sameish[0], cons[1])

        self.check_pretty_bound_constraints(
            cons[1], m.x1, {m.d1: 2, m.d2: 3, m.d3: 10, m.d4: 8}, lb=False
        )
        self.check_pretty_bound_constraints(
            cons[0], m.x1, {m.d1: 0.5, m.d2: 0.65, m.d3: 2, m.d4: -10}, lb=True
        )

        # and for x2:
        cons = mbm.get_transformed_constraints(m.d1.x2_bounds)
        self.assertEqual(len(cons), 2)
        sameish = mbm.get_transformed_constraints(m.d4.x2_lb)
        self.assertEqual(len(sameish), 1)
        self.assertIs(sameish[0], cons[0])

        self.check_pretty_bound_constraints(
            cons[1], m.x2, {m.d1: 3, m.d2: 10, m.d3: 1, m.d4: 20}, lb=False
        )
        self.check_pretty_bound_constraints(
            cons[0], m.x2, {m.d1: 0.75, m.d2: 3, m.d3: 0.55, m.d4: -5}, lb=True
        )

    def test_nested_gdp_error(self):
        m = self.make_model()
        m.d1.disjunction = Disjunction(expr=[m.x1 >= 5, m.x1 <= 4])
        with self.assertRaisesRegex(
            GDP_Error,
            "Found nested Disjunction 'd1.disjunction'. The multiple bigm "
            "transformation does not support nested GDPs. "
            "Please flatten the model before calling the "
            "transformation",
        ):
            TransformationFactory('gdp.mbigm').apply_to(m)

    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    def test_logical_constraints_on_disjuncts(self):
        m = self.make_model()
        m.d1.Y = BooleanVar()
        m.d1.Z = BooleanVar()
        m.d1.logical = LogicalConstraint(expr=m.d1.Y.implies(m.d1.Z))

        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m, bigM=self.get_Ms(m), reduce_bound_constraints=False)

        y = m.d1.Y.get_associated_binary()
        z = m.d1.Z.get_associated_binary()
        z1 = m.d1._logical_to_disjunctive.auxiliary_vars[3]

        # MbigM transformation of: (1 - z1) + (1 - y) + z >= 1
        # (1 - z1) + (1 - y) + z >= 1 - d2.ind_var - d3.ind_var
        transformed = mbm.get_transformed_constraints(
            m.d1._logical_to_disjunctive.transformed_constraints[1]
        )
        self.assertEqual(len(transformed), 1)
        c = transformed[0]
        check_obj_in_active_tree(self, c)
        self.assertIsNone(c.lower)
        self.assertEqual(value(c.upper), 0)
        repn = generate_standard_repn(c.body)
        self.assertTrue(repn.is_linear())
        simplified = repn.constant + sum(
            repn.linear_coefs[i] * repn.linear_vars[i]
            for i in range(len(repn.linear_vars))
        )
        assertExpressionsStructurallyEqual(
            self,
            simplified,
            -m.d2.binary_indicator_var - m.d3.binary_indicator_var + z1 + y - z - 1,
        )

        # MbigM transformation of: z1 + 1 - (1 - y) >= 1
        # z1 + y >= 1 - d2.ind_var - d3.ind_var
        transformed = mbm.get_transformed_constraints(
            m.d1._logical_to_disjunctive.transformed_constraints[2]
        )
        self.assertEqual(len(transformed), 1)
        c = transformed[0]
        check_obj_in_active_tree(self, c)
        self.assertIsNone(c.lower)
        self.assertEqual(value(c.upper), 0)
        repn = generate_standard_repn(c.body)
        self.assertTrue(repn.is_linear())
        simplified = repn.constant + sum(
            repn.linear_coefs[i] * repn.linear_vars[i]
            for i in range(len(repn.linear_vars))
        )
        assertExpressionsStructurallyEqual(
            self,
            simplified,
            -m.d2.binary_indicator_var - m.d3.binary_indicator_var - y - z1 + 1,
        )

        # MbigM transformation of: z1 + 1 - z >= 1
        # z1 + 1 - z >= 1 - d2.ind_var - d3.ind_var
        transformed = mbm.get_transformed_constraints(
            m.d1._logical_to_disjunctive.transformed_constraints[3]
        )
        self.assertEqual(len(transformed), 1)
        c = transformed[0]
        check_obj_in_active_tree(self, c)
        self.assertIsNone(c.lower)
        self.assertEqual(value(c.upper), 0)
        repn = generate_standard_repn(c.body)
        self.assertTrue(repn.is_linear())
        simplified = repn.constant + sum(
            repn.linear_coefs[i] * repn.linear_vars[i]
            for i in range(len(repn.linear_vars))
        )
        assertExpressionsStructurallyEqual(
            self,
            simplified,
            -m.d2.binary_indicator_var - m.d3.binary_indicator_var + z - z1,
        )

    def check_traditionally_bigmed_constraints(self, m, mbm, Ms):
        cons = mbm.get_transformed_constraints(m.d1.func)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        ub = cons[1]
        assertExpressionsEqual(
            self,
            lb.expr,
            0.0 <= m.x1 + m.x2 - m.d - Ms[m.d1][0] * (1 - m.d1.binary_indicator_var),
        )
        # [ESJ 11/23/22]: It's really hard to use assertExpressionsEqual on the
        # ub constraints because SumExpressions are sharing args, I think. So
        # when they get constructed in the transformation (because they come
        # after the lb constraints), there are nested SumExpressions. Instead of
        # trying to reproduce them I am just building a "flat" SumExpression
        # with generate_standard_repn and comparing that.
        self.assertIsNone(ub.lower)
        self.assertEqual(ub.upper, 0)
        repn = generate_standard_repn(ub.body)
        self.assertTrue(repn.is_linear())
        simplified = repn.constant + sum(
            repn.linear_coefs[i] * repn.linear_vars[i]
            for i in range(len(repn.linear_vars))
        )
        assertExpressionsEqual(
            self,
            simplified,
            m.x1 + m.x2 - m.d + Ms[m.d1][1] * m.d1.binary_indicator_var - Ms[m.d1][1],
        )

        cons = mbm.get_transformed_constraints(m.d2.func)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        ub = cons[1]
        assertExpressionsEqual(
            self,
            lb.expr,
            0.0
            <= 2 * m.x1
            + 4 * m.x2
            + 7
            - m.d
            - Ms[m.d2][0] * (1 - m.d2.binary_indicator_var),
        )
        self.assertIsNone(ub.lower)
        self.assertEqual(ub.upper, 0)
        repn = generate_standard_repn(ub.body)
        self.assertTrue(repn.is_linear())
        simplified = repn.constant + sum(
            repn.linear_coefs[i] * repn.linear_vars[i]
            for i in range(len(repn.linear_vars))
        )
        assertExpressionsEqual(
            self,
            simplified,
            2 * m.x1
            + 4 * m.x2
            - m.d
            + Ms[m.d2][1] * m.d2.binary_indicator_var
            - (Ms[m.d2][1] - 7),
        )

        cons = mbm.get_transformed_constraints(m.d3.func)
        self.assertEqual(len(cons), 2)
        lb = cons[0]
        ub = cons[1]
        assertExpressionsEqual(
            self,
            lb.expr,
            0.0
            <= m.x1
            - 5 * m.x2
            - 3
            - m.d
            - Ms[m.d3][0] * (1 - m.d3.binary_indicator_var),
        )
        self.assertIsNone(ub.lower)
        self.assertEqual(ub.upper, 0)
        repn = generate_standard_repn(ub.body)
        self.assertTrue(repn.is_linear())
        simplified = repn.constant + sum(
            repn.linear_coefs[i] * repn.linear_vars[i]
            for i in range(len(repn.linear_vars))
        )
        assertExpressionsEqual(
            self,
            simplified,
            m.x1
            - 5 * m.x2
            - m.d
            + Ms[m.d3][1] * m.d3.binary_indicator_var
            - (Ms[m.d3][1] + 3),
        )

    def test_only_multiple_bigm_bound_constraints(self):
        m = self.make_model()
        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m, only_mbigm_bound_constraints=True)

        cons = mbm.get_transformed_constraints(m.d1.x1_bounds)
        self.assertEqual(len(cons), 2)
        self.check_pretty_bound_constraints(
            cons[0], m.x1, {m.d1: 0.5, m.d2: 0.65, m.d3: 2}, lb=True
        )
        self.check_pretty_bound_constraints(
            cons[1], m.x1, {m.d1: 2, m.d2: 3, m.d3: 10}, lb=False
        )

        cons = mbm.get_transformed_constraints(m.d1.x2_bounds)
        self.assertEqual(len(cons), 2)
        self.check_pretty_bound_constraints(
            cons[0], m.x2, {m.d1: 0.75, m.d2: 3, m.d3: 0.55}, lb=True
        )
        self.check_pretty_bound_constraints(
            cons[1], m.x2, {m.d1: 3, m.d2: 10, m.d3: 1}, lb=False
        )

        self.check_traditionally_bigmed_constraints(
            m,
            mbm,
            {m.d1: (-1030.0, 1030.0), m.d2: (-1093.0, 1107.0), m.d3: (-1113.0, 1107.0)},
        )

    def test_only_multiple_bigm_bound_constraints_arg_Ms(self):
        m = self.make_model()
        mbm = TransformationFactory('gdp.mbigm')
        Ms = {m.d1: 1050, m.d2.func: (-2000, 1200), None: 4000}
        mbm.apply_to(m, only_mbigm_bound_constraints=True, bigM=Ms)

        cons = mbm.get_transformed_constraints(m.d1.x1_bounds)
        self.assertEqual(len(cons), 2)
        self.check_pretty_bound_constraints(
            cons[0], m.x1, {m.d1: 0.5, m.d2: 0.65, m.d3: 2}, lb=True
        )
        self.check_pretty_bound_constraints(
            cons[1], m.x1, {m.d1: 2, m.d2: 3, m.d3: 10}, lb=False
        )

        cons = mbm.get_transformed_constraints(m.d1.x2_bounds)
        self.assertEqual(len(cons), 2)
        self.check_pretty_bound_constraints(
            cons[0], m.x2, {m.d1: 0.75, m.d2: 3, m.d3: 0.55}, lb=True
        )
        self.check_pretty_bound_constraints(
            cons[1], m.x2, {m.d1: 3, m.d2: 10, m.d3: 1}, lb=False
        )

        self.check_traditionally_bigmed_constraints(
            m, mbm, {m.d1: (-1050, 1050), m.d2: (-2000, 1200), m.d3: (-4000, 4000)}
        )

    # A set of tests identical to test_calculated_Ms_correct, except
    # that we use each possible process spawning method for
    # multiprocessing
    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    @unittest.skipUnless(dill_available, "Dill is not available")
    def test_calculated_Ms_spawn(self):
        m = self.make_model()
        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(
            m, reduce_bound_constraints=False, threads=3, process_start_method='spawn'
        )

        self.check_all_untightened_bounds_constraints(m, mbm)
        self.check_linear_func_constraints(m, mbm)

        self.assertStructuredAlmostEqual(mbm.get_all_M_values(m), self.get_Ms(m))

    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    def test_calculated_Ms_singlethreaded(self):
        m = self.make_model()
        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m, reduce_bound_constraints=False, threads=1)

        self.check_all_untightened_bounds_constraints(m, mbm)
        self.check_linear_func_constraints(m, mbm)

        self.assertStructuredAlmostEqual(mbm.get_all_M_values(m), self.get_Ms(m))

    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    @unittest.skipUnless(dill_available, "Dill is not available")
    @unittest.skipIf(os.name == 'nt', "'forkserver' is not available on Windows")
    def test_calculated_Ms_forkserver(self):
        m = self.make_model()
        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(
            m,
            reduce_bound_constraints=False,
            threads=3,
            process_start_method='forkserver',
        )

        self.check_all_untightened_bounds_constraints(m, mbm)
        self.check_linear_func_constraints(m, mbm)

        self.assertStructuredAlmostEqual(mbm.get_all_M_values(m), self.get_Ms(m))

    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    @unittest.skipIf(os.name == 'nt', "'fork' is not available on Windows")
    def test_calculated_Ms_fork(self):
        m = self.make_model()
        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(
            m, reduce_bound_constraints=False, threads=3, process_start_method='fork'
        )

        self.check_all_untightened_bounds_constraints(m, mbm)
        self.check_linear_func_constraints(m, mbm)

        self.assertStructuredAlmostEqual(mbm.get_all_M_values(m), self.get_Ms(m))

    # Make sure we don't choke on a LegacySolverWrapper
    @unittest.skipUnless(gurobi_direct_available, "Gurobi direct is not available")
    @unittest.skipUnless(dill_available, "Dill is not available")
    def test_calculated_Ms_legacy_solver_wrapper(self):
        m = self.make_model()
        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(
            m,
            reduce_bound_constraints=False,
            threads=3,
            process_start_method='spawn',
            solver=SolverFactory('gurobi_direct_v2'),
        )

        self.check_all_untightened_bounds_constraints(m, mbm)
        self.check_linear_func_constraints(m, mbm)

        self.assertStructuredAlmostEqual(mbm.get_all_M_values(m), self.get_Ms(m))


@unittest.skipUnless(gurobi_available, "Gurobi is not available")
class NestedDisjunctsInFlatGDP(unittest.TestCase):
    """
    This class tests the fix for #2702
    """

    def test_declare_disjuncts_in_disjunction_rule(self):
        check_nested_disjuncts_in_flat_gdp(self, 'bigm')


class IndexedDisjunctiveConstraints(CommonTests):
    def test_empty_constraint_container_on_Disjunct(self):
        m = ConcreteModel()
        m.d = Disjunct()
        m.e = Disjunct()
        m.d.c = Constraint(['s', 'i', 'l', 'L', 'y'])
        m.x = Var(bounds=(2, 3))
        m.e.c = Constraint(expr=m.x == 2.7)
        m.disjunction = Disjunction(expr=[m.d, m.e])

        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m)

        cons = mbm.get_transformed_constraints(m.e.c)
        self.assertEqual(len(cons), 2)
        self.check_pretty_bound_constraints(cons[0], m.x, {m.d: 2, m.e: 2.7}, lb=True)
        self.check_pretty_bound_constraints(cons[1], m.x, {m.d: 3, m.e: 2.7}, lb=False)


@unittest.skipUnless(gurobi_available, "Gurobi is not available")
class IndexedDisjunction(unittest.TestCase):
    def test_two_term_indexed_disjunction(self):
        """
        This test checks that we don't do anything silly with transformation Blocks in
        the case that the Disjunction is indexed.
        """
        m = make_indexed_equality_model()

        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m)

        cons = mbm.get_transformed_constraints(m.d[1].disjuncts[0].constraint[1])
        self.assertEqual(len(cons), 2)
        assertExpressionsEqual(
            self,
            cons[0].expr,
            m.x[1]
            >= m.d[1].disjuncts[0].binary_indicator_var
            + 2.0 * m.d[1].disjuncts[1].binary_indicator_var,
        )
        assertExpressionsEqual(
            self,
            cons[1].expr,
            m.x[1]
            <= m.d[1].disjuncts[0].binary_indicator_var
            + 2.0 * m.d[1].disjuncts[1].binary_indicator_var,
        )
        cons_again = mbm.get_transformed_constraints(m.d[1].disjuncts[1].constraint[1])
        self.assertEqual(len(cons_again), 2)
        self.assertIs(cons_again[0], cons[0])
        self.assertIs(cons_again[1], cons[1])

        cons = mbm.get_transformed_constraints(m.d[2].disjuncts[0].constraint[1])
        self.assertEqual(len(cons), 2)
        assertExpressionsEqual(
            self,
            cons[0].expr,
            m.x[2]
            >= m.d[2].disjuncts[0].binary_indicator_var
            + 2.0 * m.d[2].disjuncts[1].binary_indicator_var,
        )
        assertExpressionsEqual(
            self,
            cons[1].expr,
            m.x[2]
            <= m.d[2].disjuncts[0].binary_indicator_var
            + 2.0 * m.d[2].disjuncts[1].binary_indicator_var,
        )
        cons_again = mbm.get_transformed_constraints(m.d[2].disjuncts[1].constraint[1])
        self.assertEqual(len(cons_again), 2)
        self.assertIs(cons_again[0], cons[0])
        self.assertIs(cons_again[1], cons[1])


class EdgeCases(unittest.TestCase):
    def make_infeasible_disjunct_model(self):
        m = ConcreteModel()
        m.x = Var(bounds=(1, 12))
        m.y = Var(bounds=(19, 22))
        m.disjunction = Disjunction(
            expr=[
                [m.x >= 3 + m.y, m.y == 19.75],  # infeasible given bounds
                [m.y >= 21 + m.x],  # unique solution
                [m.x == m.y - 9],  # x in interval [10, 12]
            ]
        )
        return m

    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    def test_calculate_Ms_infeasible_Disjunct(self):
        m = self.make_infeasible_disjunct_model()
        out = StringIO()
        mbm = TransformationFactory('gdp.mbigm')
        with LoggingIntercept(out, 'pyomo.gdp.mbigm', logging.DEBUG):
            mbm.apply_to(m, reduce_bound_constraints=False)

        # We mentioned the infeasibility at the DEBUG level
        self.assertIn(
            r"Disjunct 'disjunction_disjuncts[0]' is infeasible, deactivating",
            out.getvalue().strip(),
        )

        # We just fixed the infeasible disjunct to False
        self.assertFalse(m.disjunction.disjuncts[0].active)
        self.assertTrue(m.disjunction.disjuncts[0].indicator_var.fixed)
        self.assertFalse(value(m.disjunction.disjuncts[0].indicator_var))

        # We didn't actually transform the infeasible disjunct
        self.assertIsNone(m.disjunction.disjuncts[0].transformation_block)

        # the remaining constraints are transformed correctly.
        cons = mbm.get_transformed_constraints(m.disjunction.disjuncts[1].constraint[1])
        self.assertEqual(len(cons), 1)
        assertExpressionsEqual(
            self,
            cons[0].expr,
            21 + m.x - m.y <= 12.0 * m.disjunction.disjuncts[2].binary_indicator_var,
        )

        cons = mbm.get_transformed_constraints(m.disjunction.disjuncts[2].constraint[1])
        self.assertEqual(len(cons), 2)
        assertExpressionsEqual(
            self,
            cons[0].expr,
            -12.0 * m.disjunction_disjuncts[1].binary_indicator_var <= m.x - (m.y - 9),
        )
        assertExpressionsEqual(
            self,
            cons[1].expr,
            m.x - (m.y - 9) <= -12.0 * m.disjunction_disjuncts[1].binary_indicator_var,
        )

    @unittest.skipUnless(
        SolverFactory('ipopt').available(exception_flag=False), "Ipopt is not available"
    )
    def test_calculate_Ms_infeasible_Disjunct_local_solver(self):
        m = self.make_infeasible_disjunct_model()
        # When multiple exceptions are raised during a
        # multiprocessing.Pool.map call, it is indeterminate which
        # exception will be raised to the caller.
        with self.assertRaisesRegex(
            GDP_Error,
            r"Unsuccessful solve to calculate M value to "
            r"relax constraint 'disjunction_disjuncts\[\d+\].constraint\[\d+\]' "
            r"on Disjunct 'disjunction_disjuncts\[\d+\]' when "
            r"Disjunct 'disjunction_disjuncts\[\d+\]' is selected.",
        ):
            TransformationFactory('gdp.mbigm').apply_to(
                m,
                solver=SolverFactory('ipopt'),
                reduce_bound_constraints=False,
                use_primal_bound=True,
            )

    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    def test_politely_ignore_BigM_Suffix(self):
        m = self.make_infeasible_disjunct_model()
        m.disjunction.disjuncts[0].deactivate()
        m.disjunction.disjuncts[1].BigM = Suffix(direction=Suffix.LOCAL)
        out = StringIO()
        with LoggingIntercept(out, 'pyomo.gdp.mbigm', logging.DEBUG):
            TransformationFactory('gdp.mbigm').apply_to(
                m, reduce_bound_constraints=False
            )
        warnings = out.getvalue()
        self.assertIn(
            r"Found active 'BigM' Suffix on 'disjunction_disjuncts[1]'. "
            r"The multiple bigM transformation does not currently "
            r"support specifying M's with Suffixes and is ignoring "
            r"this Suffix.",
            warnings,
        )

    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    def test_complain_for_unrecognized_Suffix(self):
        m = self.make_infeasible_disjunct_model()
        m.disjunction.disjuncts[0].deactivate()
        m.disjunction.disjuncts[1].HiThere = Suffix(direction=Suffix.LOCAL)
        out = StringIO()
        with self.assertRaisesRegex(
            GDP_Error,
            r"Found active Suffix 'disjunction_disjuncts\[1\].HiThere' "
            r"on Disjunct 'disjunction_disjuncts\[1\]'. The multiple bigM "
            r"transformation does not support this Suffix.",
        ):
            TransformationFactory('gdp.mbigm').apply_to(
                m, reduce_bound_constraints=False
            )

    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    def test_transform_on_block(self):
        # In multiple_bigm.py, if constraint.getname and disjunct.getname do not
        # set relative_to=instance, the transformation fails because of mismatched
        # names in _calc_M.
        # This test ensures that the transformation works on blocks as well as
        # ConcreteModels.
        m = ConcreteModel()
        m.b = Block()
        m.b.x = Var(bounds=(0, 5))
        m.b.y = Var()
        m.b.dis1 = Disjunct()
        m.b.dis2 = Disjunct()

        m.b.dis1.linear = Constraint(expr=m.b.x * 0.5 + 3 == m.b.y)
        m.b.dis2.linear = Constraint(expr=m.b.x * 0.2 + 1 == m.b.y)
        m.b.d = Disjunction(expr=[m.b.dis1, m.b.dis2])

        mbm = TransformationFactory('gdp.mbigm')
        mbm.apply_to(m.b, threads=1)
        dis1_cons = mbm.get_transformed_constraints(m.b.dis1.linear)
        assertExpressionsEqual(
            self,
            dis1_cons[0].expr,
            2.0 * m.b.dis2.binary_indicator_var <= 0.5 * m.b.x + 3 - m.b.y,
        )
        assertExpressionsEqual(
            self,
            dis1_cons[1].expr,
            0.5 * m.b.x + 3 - m.b.y <= 3.5 * m.b.dis2.binary_indicator_var,
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Re-implementation of example 1 of Quesada and Grossmann.

Re-implementation of Quesada example 2 MINLP test problem in Pyomo
Author: David Bernal <https://github.com/bernalde>.

The expected optimal solution value is -5.512.

Ref:
    Quesada, Ignacio, and Ignacio E. Grossmann.
    'An LP/NLP based branch and bound algorithm
    for convex MINLP optimization problems.'
    Computers & chemical engineering 16.10-11 (1992): 937-947.

    Problem type:    convex MINLP
            size:    1  binary variable
                     2  continuous variables
                     4  constraints


"""

from pyomo.environ import (
    Binary,
    ConcreteModel,
    Constraint,
    Reals,
    Objective,
    RangeSet,
    Var,
    minimize,
    log,
)
from pyomo.common.collections import ComponentMap


class SimpleMINLP(ConcreteModel):
    def __init__(self, *args, **kwargs):
        """Create the problem."""
        kwargs.setdefault('name', 'SimpleMINLP3')
        super(SimpleMINLP, self).__init__(*args, **kwargs)
        m = self

        """Set declarations"""
        I = m.I = RangeSet(1, 2, doc='continuous variables')
        J = m.J = RangeSet(1, 1, doc='discrete variables')

        # initial point information for discrete variables
        initY = {1: 1}
        # initial point information for continuous variables
        initX = {1: 0, 2: 0}

        """Variable declarations"""
        # DISCRETE VARIABLES
        Y = m.Y = Var(J, domain=Binary, initialize=initY)
        # CONTINUOUS VARIABLES
        X = m.X = Var(I, domain=Reals, initialize=initX, bounds=(-0.9, 50))

        """Constraint definitions"""
        # CONSTRAINTS
        m.const1 = Constraint(expr=-X[2] + 5 * log(X[1] + 1) + 3 * Y[1] >= 0)
        m.const2 = Constraint(expr=-X[2] + X[1] ** 2 - Y[1] <= 1)
        m.const3 = Constraint(expr=X[1] + X[2] + 20 * Y[1] <= 24)
        m.const4 = Constraint(expr=2 * X[2] + 3 * X[1] <= 10)

        """Cost (objective) function definition"""
        m.objective = Objective(
            expr=10 * X[1] ** 2 - X[2] + 5 * (Y[1] - 1), sense=minimize
        )
        m.optimal_value = -5.512
        m.optimal_solution = ComponentMap()
        m.optimal_solution[m.X[1]] = 0.20710677582302733
        m.optimal_solution[m.X[2]] = 0.9411320859243828
        m.optimal_solution[m.Y[1]] = 0.0



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo

from pyomo.contrib.pynumero.dependencies import (
    numpy as np,
    numpy_available,
    scipy,
    scipy_available,
)
from pyomo.common.dependencies.scipy import sparse as spa

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pynumero needs scipy and numpy to run NLP tests")

from pyomo.contrib.pynumero.asl import AmplInterface

if not AmplInterface.available():
    raise unittest.SkipTest(
        "Pynumero needs the ASL extension to run CyIpoptSolver tests"
    )

from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP

from pyomo.contrib.pynumero.interfaces.cyipopt_interface import cyipopt_available

if not cyipopt_available:
    raise unittest.SkipTest("Pynumero needs cyipopt to run CyIpoptSolver tests")

from pyomo.contrib.pynumero.interfaces.cyipopt_interface import CyIpoptNLP


def create_model1():
    m = pyo.ConcreteModel()
    m.x = pyo.Var([1, 2, 3], initialize=4.0)
    m.d = pyo.Constraint(expr=m.x[1] + m.x[2] ** 2 <= 18.0)
    m.c = pyo.Constraint(expr=m.x[3] ** 2 + m.x[1] == 25)
    m.o = pyo.Objective(expr=m.x[1] ** 4 - 3 * m.x[1] * m.x[2] ** 3 + m.x[3] ** 2 - 8.0)
    m.x[1].setlb(0.0)
    m.x[2].setlb(0.0)
    return m


class TestCyIpoptNLP(unittest.TestCase):
    def test_model1_CyIpoptNLP(self):
        model = create_model1()
        nlp = PyomoNLP(model)
        cynlp = CyIpoptNLP(nlp)
        self._check_model1(nlp, cynlp)

    def test_model1_CyIpoptNLP_scaling(self):
        m = create_model1()
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.scaling_factor[m.o] = 1e-6  # scale the objective
        m.scaling_factor[m.c] = 2.0  # scale the equality constraint
        m.scaling_factor[m.d] = 3.0  # scale the inequality constraint
        m.scaling_factor[m.x[1]] = 4.0  # scale one of the x variables

        cynlp = CyIpoptNLP(PyomoNLP(m))
        obj_scaling, x_scaling, g_scaling = cynlp.scaling_factors()
        self.assertTrue(obj_scaling == 1e-6)
        self.assertTrue(len(x_scaling) == 3)
        # vars are in order x[2], x[3], x[1]
        self.assertTrue(x_scaling[0] == 1.0)
        self.assertTrue(x_scaling[1] == 1.0)
        self.assertTrue(x_scaling[2] == 4.0)
        self.assertTrue(len(g_scaling) == 2)
        # assuming the order is d then c
        self.assertTrue(g_scaling[0] == 3.0)
        self.assertTrue(g_scaling[1] == 2.0)

        # test missing obj
        m = create_model1()
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        # m.scaling_factor[m.o] = 1e-6 # scale the objective
        m.scaling_factor[m.c] = 2.0  # scale the equality constraint
        m.scaling_factor[m.d] = 3.0  # scale the inequality constraint
        m.scaling_factor[m.x[1]] = 4.0  # scale the x variable

        cynlp = CyIpoptNLP(PyomoNLP(m))
        obj_scaling, x_scaling, g_scaling = cynlp.scaling_factors()
        self.assertTrue(obj_scaling == 1.0)
        self.assertTrue(len(x_scaling) == 3)
        # vars are in order x[2], x[3], x[1]
        self.assertTrue(x_scaling[0] == 1.0)
        self.assertTrue(x_scaling[1] == 1.0)
        self.assertTrue(x_scaling[2] == 4.0)
        self.assertTrue(len(g_scaling) == 2)
        # assuming the order is d then c
        self.assertTrue(g_scaling[0] == 3.0)
        self.assertTrue(g_scaling[1] == 2.0)

        # test missing var
        m = create_model1()
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.scaling_factor[m.o] = 1e-6  # scale the objective
        m.scaling_factor[m.c] = 2.0  # scale the equality constraint
        m.scaling_factor[m.d] = 3.0  # scale the inequality constraint
        # m.scaling_factor[m.x] = 4.0  # scale the x variable

        cynlp = CyIpoptNLP(PyomoNLP(m))
        obj_scaling, x_scaling, g_scaling = cynlp.scaling_factors()
        self.assertTrue(obj_scaling == 1e-6)
        self.assertTrue(len(x_scaling) == 3)
        # vars are in order x[2], x[3], x[1]
        self.assertTrue(x_scaling[0] == 1.0)
        self.assertTrue(x_scaling[1] == 1.0)
        self.assertTrue(x_scaling[2] == 1.0)
        self.assertTrue(len(g_scaling) == 2)
        # assuming the order is d then c
        self.assertTrue(g_scaling[0] == 3.0)
        self.assertTrue(g_scaling[1] == 2.0)

        # test missing c
        m = create_model1()
        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)

        m.scaling_factor[m.o] = 1e-6  # scale the objective
        # m.scaling_factor[m.c] = 2.0  # scale the equality constraint
        m.scaling_factor[m.d] = 3.0  # scale the inequality constraint
        m.scaling_factor[m.x[1]] = 4.0  # scale the x variable

        cynlp = CyIpoptNLP(PyomoNLP(m))
        obj_scaling, x_scaling, g_scaling = cynlp.scaling_factors()
        self.assertTrue(obj_scaling == 1e-6)
        self.assertTrue(len(x_scaling) == 3)
        # vars are in order x[2], x[3], x[1]
        self.assertTrue(x_scaling[0] == 1.0)
        self.assertTrue(x_scaling[1] == 1.0)
        self.assertTrue(x_scaling[2] == 4.0)
        self.assertTrue(len(g_scaling) == 2)
        # assuming the order is d then c
        self.assertTrue(g_scaling[0] == 3.0)
        self.assertTrue(g_scaling[1] == 1.0)

        # test missing all
        m = create_model1()
        # m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        # m.scaling_factor[m.o] = 1e-6 # scale the objective
        # m.scaling_factor[m.c] = 2.0  # scale the equality constraint
        # m.scaling_factor[m.d] = 3.0  # scale the inequality constraint
        # m.scaling_factor[m.x] = 4.0  # scale the x variable

        cynlp = CyIpoptNLP(PyomoNLP(m))
        obj_scaling, x_scaling, g_scaling = cynlp.scaling_factors()
        self.assertTrue(obj_scaling is None)
        self.assertTrue(x_scaling is None)
        self.assertTrue(g_scaling is None)

    def _check_model1(self, nlp, cynlp):
        # test x_init
        expected_xinit = np.asarray([4.0, 4.0, 4.0], dtype=np.float64)
        xinit = cynlp.x_init()
        self.assertTrue(np.array_equal(xinit, expected_xinit))

        # test x_lb
        expected_xlb = list()
        for v in nlp.get_pyomo_variables():
            if v.lb == None:
                expected_xlb.append(-np.inf)
            else:
                expected_xlb.append(v.lb)
        expected_xlb = np.asarray(expected_xlb)
        xlb = cynlp.x_lb()
        self.assertTrue(np.array_equal(xlb, expected_xlb))

        # test x_ub
        expected_xub = list()
        for v in nlp.get_pyomo_variables():
            if v.ub == None:
                expected_xub.append(np.inf)
            else:
                expected_xub.append(v.ub)
        expected_xub = np.asarray(expected_xub)
        xub = cynlp.x_ub()
        self.assertTrue(np.array_equal(xub, expected_xub))

        # test g_lb
        expected_glb = np.asarray([-np.inf, 0.0], dtype=np.float64)
        glb = cynlp.g_lb()
        self.assertTrue(np.array_equal(glb, expected_glb))

        # test g_ub
        expected_gub = np.asarray([18, 0.0], dtype=np.float64)
        gub = cynlp.g_ub()
        print(expected_gub)
        print(gub)
        self.assertTrue(np.array_equal(gub, expected_gub))

        x = cynlp.x_init()
        # test objective
        self.assertEqual(cynlp.objective(x), -504)
        # test gradient
        expected = np.asarray([-576, 8, 64], dtype=np.float64)
        self.assertTrue(np.allclose(expected, cynlp.gradient(x)))
        # test constraints
        expected = np.asarray([20, -5], dtype=np.float64)
        constraints = cynlp.constraints(x)
        self.assertTrue(np.allclose(expected, constraints))

        # test jacobian
        expected = np.asarray([[8.0, 0, 1.0], [0.0, 8.0, 1.0]])
        spexpected = spa.coo_matrix(expected).todense()
        rows, cols = cynlp.jacobianstructure()
        values = cynlp.jacobian(x)
        jac = spa.coo_matrix(
            (values, (rows, cols)), shape=(len(constraints), len(x))
        ).todense()
        self.assertTrue(np.allclose(spexpected, jac))

        # test hessian
        y = constraints.copy()
        y.fill(1.0)
        rows, cols = cynlp.hessianstructure()
        values = cynlp.hessian(x, y, obj_factor=1.0)
        hess_lower = spa.coo_matrix(
            (values, (rows, cols)), shape=(len(x), len(x))
        ).todense()
        expected_hess_lower = np.asarray(
            [[-286.0, 0.0, 0.0], [0.0, 4.0, 0.0], [-144.0, 0.0, 192.0]],
            dtype=np.float64,
        )
        self.assertTrue(np.allclose(expected_hess_lower, hess_lower))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import math

from pyomo.core.staleflag import StaleFlagManager
from pyomo.core.expr.symbol_map import SymbolMap
from pyomo.core.kernel.base import _no_ctype, _convert_ctype
from pyomo.core.kernel.heterogeneous_container import IHeterogeneousContainer
from pyomo.core.kernel.container_utils import define_simple_containers

logger = logging.getLogger('pyomo.core')


class IBlock(IHeterogeneousContainer):
    """A generalized container that can store objects of
    any category type as attributes.
    """

    __slots__ = ()
    _child_storage_delimiter_string = "."
    _child_storage_entry_string = "%s"

    #
    # Define the IHeterogeneousContainer abstract methods
    #

    # def child_ctypes(self, *args, **kwds):
    # ... not defined here

    #
    # Define the ICategorizedObjectContainer abstract methods
    #

    def child(self, key):
        """Get the child object associated with a given
        storage key for this container.

        Raises:
            KeyError: if the argument is not a storage key
                for any children of this container
        """
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(str(key))

    # def children(self, *args, **kwds):
    # ... not defined here


class block(IBlock):
    """A generalized container for defining hierarchical
    models by adding modeling components as attributes.

    Examples:
        >>> import pyomo.kernel as pmo
        >>> model = pmo.block()
        >>> model.x = pmo.variable()
        >>> model.c = pmo.constraint(model.x >= 1)
        >>> model.o = pmo.objective(model.x)
    """

    _ctype = IBlock
    # determines when large block storage is activated
    _lbs_count = 4
    _block_reserved_words = set()

    @staticmethod
    def _refresh_block_reserved_words():
        block._block_reserved_words = set(dir(block()))
        # this is protected because it is
        # defined as a property with a setter
        # that explicitly disallows this
        block._block_reserved_words.remove('active')

    def __init__(self):
        # bypass __setatrr__ for this class
        d = self.__dict__
        d['_parent'] = None
        d['_storage_key'] = None
        d['_active'] = True
        d['_block__byctype'] = None
        d['_block__order'] = dict()

    def _activate_large_storage_mode(self):
        if self.__byctype.__class__ is not dict:
            self_byctype = self.__dict__['_block__byctype'] = dict()
            for key, obj in self.__order.items():
                ctype = obj.ctype
                if ctype not in self_byctype:
                    self_byctype[ctype] = dict()
                self_byctype[ctype][key] = obj

    #
    # Define the IHeterogeneousContainer abstract methods
    #

    def child_ctypes(self):
        """Returns the set of child object category types
        stored in this container."""
        self_byctype = self.__byctype
        if self_byctype is None:
            # empty
            return ()
        elif self_byctype.__class__ is int:
            # small block storage
            # (self_byctype is a union of hash bytes)
            ctypes_set = set()
            ctypes = []
            for child in self.__order.values():
                child_ctype = child.ctype
                if child_ctype not in ctypes_set:
                    ctypes_set.add(child_ctype)
                    ctypes.append(child_ctype)
            return tuple(ctypes)
        elif self_byctype.__class__ is dict:
            # large-block storage
            return tuple(self_byctype)
        else:
            # storing a single ctype
            return (self_byctype,)

    #
    # Define the ICategorizedObjectContainer abstract methods
    #

    def children(self, ctype=_no_ctype):
        """Iterate over the children of this block.

        Args:
            ctype: Indicates the category of children to
                include. The default value indicates that
                all categories should be included.

        Returns:
            iterator of child objects
        """
        self_byctype = self.__byctype
        if self_byctype is None:
            # empty
            return

        # convert AML types into Kernel types (hack for the
        # solver interfaces)
        ctype = _convert_ctype.get(ctype, ctype)

        if ctype is _no_ctype:
            yield from self.__order.values()
        elif self_byctype.__class__ is dict:
            # large-block storage
            if ctype in self_byctype:
                yield from self_byctype[ctype].values()
        elif self_byctype.__class__ is int:
            # small-block storage
            # (self_byctype is a union of hash bytes)
            h_ = hash(ctype)
            if (self_byctype & h_) == h_:
                for child in self.__order.values():
                    if child.ctype is ctype:
                        yield child
        elif self_byctype is ctype:
            # storing a single ctype
            yield from self.__order.values()

    #
    # Interface
    #

    def __setattr__(self, name, obj):
        if name in self._block_reserved_words:
            raise ValueError(
                "Attempting to modify a reserved block attribute: %s" % (name,)
            )
        needs_del = False
        same_obj = False
        self_order = self.__order
        if name in self_order:
            # to avoid an edge case, we need to delay
            # deleting the current object until we
            # check the parent of the new object
            needs_del = True
            if self_order[name] is not obj:
                logger.warning(
                    "Implicitly replacing attribute %s (type=%s) "
                    "on block with new object (type=%s). This "
                    "is usually indicative of a modeling error. "
                    "To avoid this warning, delete the original "
                    "object from the block before assigning a new "
                    "object."
                    % (
                        name,
                        getattr(self, name).__class__.__name__,
                        obj.__class__.__name__,
                    )
                )
            else:
                same_obj = True
                assert obj.parent is self

        try:
            ctype = obj.ctype
        except AttributeError:
            if needs_del:
                delattr(self, name)
        else:
            if (obj.parent is None) or same_obj:
                if needs_del:
                    delattr(self, name)
                obj._update_parent_and_storage_key(self, name)
                self_order[name] = obj
                self_byctype = self.__byctype
                if self_byctype is None:
                    # storing a single ctype
                    self.__dict__['_block__byctype'] = ctype
                elif self_byctype.__class__ is dict:
                    # large-block storage
                    if ctype not in self_byctype:
                        self_byctype[ctype] = dict()
                    self_byctype[ctype][name] = obj
                elif self_byctype.__class__ is int:
                    # small-block storage
                    # (_byctype is a union of hash bytes)
                    if len(self_order) > self._lbs_count:
                        # activate the large storage format
                        # if we have exceeded the threshold
                        self._activate_large_storage_mode()
                    else:
                        self.__dict__['_block__byctype'] |= hash(ctype)
                else:
                    # currently storing a single ctype
                    if ctype is not self_byctype:
                        if len(self_order) > self._lbs_count:
                            # activate the large storage format
                            # if we have exceeded the threshold
                            self._activate_large_storage_mode()
                        else:
                            self.__dict__['_block__byctype'] = hash(
                                self_byctype
                            ) | hash(ctype)
            else:
                raise ValueError(
                    "Invalid assignment to %s type with name '%s' "
                    "at entry %s. A parent container has already "
                    "been assigned to the object being inserted: %s"
                    % (self.__class__.__name__, self.name, name, obj.parent.name)
                )
        super(block, self).__setattr__(name, obj)

    def __delattr__(self, name):
        self_order = self.__order
        if name in self_order:
            obj = self_order[name]
            del self_order[name]
            obj._clear_parent_and_storage_key()
            self_byctype = self.__byctype
            if self_byctype.__class__ is dict:
                # large-block storage
                ctype = obj.ctype
                del self_byctype[ctype][name]
                if len(self_byctype[ctype]) == 0:
                    del self_byctype[ctype]
        super(block, self).__delattr__(name)

    def write(
        self,
        filename,
        format=None,
        _solver_capability=None,
        _called_by_solver=False,
        **kwds,
    ):
        """
        Write the model to a file, with a given format.

        Args:
            filename (str): The name of the file to write.
            format: The file format to use. If this is not
                specified, the file format will be inferred
                from the filename suffix.
            **kwds: Additional keyword options passed to the
                model writer.

        Returns:
            a :class:`SymbolMap`
        """
        import pyomo.opt

        #
        # Guess the format if none is specified
        #
        if format is None:
            format = pyomo.opt.base.guess_format(filename)
        problem_writer = pyomo.opt.WriterFactory(format)

        if problem_writer is None:
            raise ValueError(
                "Cannot write model in format '%s': no model "
                "writer registered for that format" % str(format)
            )

        if _solver_capability is None:
            _solver_capability = lambda x: True
        filename_, smap = problem_writer(self, filename, _solver_capability, kwds)
        assert filename_ == filename

        if _called_by_solver:
            # BIG HACK
            smap_id = id(smap)
            if not hasattr(self, "._symbol_maps"):
                setattr(self, "._symbol_maps", {})
            getattr(self, "._symbol_maps")[smap_id] = smap
            return smap_id
        else:
            return smap

    def load_solution(
        self,
        solution,
        allow_consistent_values_for_fixed_vars=False,
        comparison_tolerance_for_fixed_vars=1e-5,
    ):
        """
        Load a solution.

        Args:
            solution: A :class:`pyomo.opt.Solution` object with a
                symbol map. Optionally, the solution can be tagged
                with a default variable value (e.g., 0) that will be
                applied to those variables in the symbol map that do
                not have a value in the solution.
            allow_consistent_values_for_fixed_vars:
                Indicates whether a solution can specify
                consistent values for variables that are
                fixed.
            comparison_tolerance_for_fixed_vars: The
                tolerance used to define whether or not a
                value in the solution is consistent with the
                value of a fixed variable.
        """
        from pyomo.core.kernel.suffix import import_suffix_generator

        symbol_map = solution.symbol_map
        default_variable_value = getattr(solution, "default_variable_value", None)

        # Generate the list of active import suffixes on
        # this top level model
        valid_import_suffixes = {
            obj.storage_key: obj for obj in import_suffix_generator(self)
        }

        # To ensure that import suffix data gets properly
        # overwritten (e.g., the case where nonzero dual
        # values exist on the suffix and but only sparse
        # dual values exist in the results object) we clear
        # all active import suffixes.
        for suffix in valid_import_suffixes.values():
            suffix.clear()

        # Load problem (model) level suffixes. These would
        # only come from ampl interfaced solution suffixes
        # at this point in time.
        for _attr_key, attr_value in solution.problem.items():
            attr_key = _attr_key[0].lower() + _attr_key[1:]
            if attr_key in valid_import_suffixes:
                valid_import_suffixes[attr_key][self] = attr_value

        #
        # Set the "stale" flag of each variable in the model prior to
        # loading the solution, so you known which variables have "real"
        # values and which ones don't.
        #
        StaleFlagManager.mark_all_as_stale()
        #
        # Load variable data
        #
        from pyomo.core.kernel.variable import IVariable

        var_skip_attrs = ['id', 'canonical_label']
        seen_var_ids = set()
        for label, entry in solution.variable.items():
            var = symbol_map.getObject(label)
            if (var is None) or (var is SymbolMap.UnknownSymbol):
                # NOTE: the following is a hack, to handle
                #    the ONE_VAR_CONSTANT variable that is
                #    necessary for the objective
                #    constant-offset terms.  probably should
                #    create a dummy variable in the model
                #    map at the same time the objective
                #    expression is being constructed.
                if "ONE_VAR_CONST" in label:
                    continue
                else:
                    raise KeyError(
                        "Variable associated with symbol '%s' "
                        "is not found on this block" % (label)
                    )

            seen_var_ids.add(id(var))

            if (not allow_consistent_values_for_fixed_vars) and var.fixed:
                raise ValueError(
                    "Variable '%s' is currently fixed. "
                    "A new value is not expected "
                    "in solution" % (var.name)
                )

            for _attr_key, attr_value in entry.items():
                attr_key = _attr_key[0].lower() + _attr_key[1:]
                if attr_key == 'value':
                    if (
                        allow_consistent_values_for_fixed_vars
                        and var.fixed
                        and (
                            math.fabs(attr_value - var.value)
                            > comparison_tolerance_for_fixed_vars
                        )
                    ):
                        raise ValueError(
                            "Variable %s is currently fixed. "
                            "A value of '%s' in solution is "
                            "not within tolerance=%s of the current "
                            "value of '%s'"
                            % (
                                var.name,
                                attr_value,
                                comparison_tolerance_for_fixed_vars,
                                var.value,
                            )
                        )
                    var.set_value(attr_value, skip_validation=True)
                elif attr_key in valid_import_suffixes:
                    valid_import_suffixes[attr_key][var] = attr_value

        # start to build up the set of unseen variable ids
        unseen_var_ids = set(symbol_map.byObject.keys())
        # at this point it contains ids for non-variable types
        unseen_var_ids.difference_update(seen_var_ids)

        #
        # Load objective solution (should simply be suffixes if
        # they exist)
        #
        objective_skip_attrs = ['id', 'canonical_label', 'value']
        for label, entry in solution.objective.items():
            obj = symbol_map.getObject(label)
            if (obj is None) or (obj is SymbolMap.UnknownSymbol):
                raise KeyError(
                    "Objective associated with symbol '%s' "
                    "is not found on this block" % (label)
                )
            # Because of __default_objective__, an objective might
            # appear twice in the objective dictionary.
            unseen_var_ids.discard(id(obj))
            for _attr_key, attr_value in entry.items():
                attr_key = _attr_key[0].lower() + _attr_key[1:]
                if attr_key in valid_import_suffixes:
                    valid_import_suffixes[attr_key][obj] = attr_value

        #
        # Load constraint solution
        #
        con_skip_attrs = ['id', 'canonical_label']
        for label, entry in solution.constraint.items():
            con = symbol_map.getObject(label)
            if con is SymbolMap.UnknownSymbol:
                #
                # This is a hack - see above.
                #
                if "ONE_VAR_CONST" in label:
                    continue
                else:
                    raise KeyError(
                        "Constraint associated with symbol '%s' "
                        "is not found on this block" % (label)
                    )
            unseen_var_ids.discard(id(con))
            for _attr_key, attr_value in entry.items():
                attr_key = _attr_key[0].lower() + _attr_key[1:]
                if attr_key in valid_import_suffixes:
                    valid_import_suffixes[attr_key][con] = attr_value

        #
        # Load sparse variable solution
        #
        if default_variable_value is not None:
            for var_id in unseen_var_ids:
                var = symbol_map.getObject(symbol_map.byObject[var_id])
                if var.ctype is not IVariable:
                    continue
                if (not allow_consistent_values_for_fixed_vars) and var.fixed:
                    raise ValueError(
                        "Variable '%s' is currently fixed. "
                        "A new value is not expected "
                        "in solution" % (var.name)
                    )

                if (
                    allow_consistent_values_for_fixed_vars
                    and var.fixed
                    and (
                        math.fabs(default_variable_value - var.value)
                        > comparison_tolerance_for_fixed_vars
                    )
                ):
                    raise ValueError(
                        "Variable %s is currently fixed. "
                        "A value of '%s' in solution is "
                        "not within tolerance=%s of the current "
                        "value of '%s'"
                        % (
                            var.name,
                            default_variable_value,
                            comparison_tolerance_for_fixed_vars,
                            var.value,
                        )
                    )
                var.set_value(default_variable_value, skip_validation=True)

        # Set the state flag to "delayed advance": it will auto-advance
        # if a non-stale variable is updated (causing all non-stale
        # variables to be marked as stale).
        StaleFlagManager.mark_all_as_stale(delayed=True)


# inserts class definitions for simple _tuple, _list, and
# _dict containers into this module
define_simple_containers(globals(), "block", IBlock)

# populate the initial set of reserved block attributes so
# that users can not overwrite them when building a model
block._refresh_block_reserved_words()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.opt.base.solvers import SolverFactory
from pyomo.common import Executable
from pyomo.common.collections import Bunch
from pyomo.solvers.plugins.solvers.ASL import ASL

logger = logging.getLogger('pyomo.solvers')


@SolverFactory.register('path', doc='Nonlinear MCP solver')
class PATHAMPL(ASL):
    """An interface to the PATH MCP solver."""

    def __init__(self, **kwds):
        #
        # Call base constructor
        #
        kwds["type"] = "path"
        ASL.__init__(self, **kwds)
        self._metasolver = False
        #
        # Define solver capabilities, which default to 'None'
        #
        self._capabilities = Bunch()
        self._capabilities.linear = True

    def _default_executable(self):
        executable = Executable("pathampl")
        if not executable:  # pragma:nocover
            logger.warning(
                "Could not locate the 'pathampl' executable, "
                "which is required for solver %s" % self.name
            )
            self.enable = False
            return None
        return executable.path()

    def create_command_line(self, executable, problem_files):
        self.options.solver = 'pathampl'
        return ASL.create_command_line(self, executable, problem_files)



# -*- coding: utf-8 -*-
# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for pyomo.base.misc
#
from io import StringIO

import pyomo.common.unittest as unittest

from pyomo.common.sorting import sorted_robust, _robust_sort_keyfcn


# The following are custom types used for testing sorted_robust.  They
# are declared at the module scope to ensure consistent generation of
# the class __name__.
class LikeFloat:
    def __init__(self, n):
        self.n = n

    def __lt__(self, other):
        return self.n < other

    def __gt__(self, other):
        return self.n > other


class Comparable:
    def __init__(self, n):
        self.n = str(n)

    def __lt__(self, other):
        return self.n < other

    def __gt__(self, other):
        return self.n > other


class ToStr:
    def __init__(self, n):
        self.n = str(n)

    def __str__(self):
        return self.n


class NoStr:
    def __init__(self, n):
        self.n = str(n)

    def __str__(self):
        raise ValueError('')


class TestSortedRobust(unittest.TestCase):
    def test_sorted_robust(self):
        # Note: as types are sorted by name, int < str < tuple
        a = sorted_robust([3, 2, 1])
        self.assertEqual(a, [1, 2, 3])

        # Testthat ints and floats are sorted as "numbers"
        a = sorted_robust([3, 2.1, 1])
        self.assertEqual(a, [1, 2.1, 3])

        a = sorted_robust([3, '2', 1])
        self.assertEqual(a, [1, 3, '2'])

        a = sorted_robust([('str1', 'str1'), (1, 'str2')])
        self.assertEqual(a, [(1, 'str2'), ('str1', 'str1')])

        a = sorted_robust([((1,), 'str2'), ('str1', 'str1')])
        self.assertEqual(a, [('str1', 'str1'), ((1,), 'str2')])

        a = sorted_robust([('str1', 'str1'), ((1,), 'str2')])
        self.assertEqual(a, [('str1', 'str1'), ((1,), 'str2')])

    def test_user_key(self):
        # ensure it doesn't throw an error
        # Test for issue https://github.com/Pyomo/pyomo/issues/2019
        sorted_robust([(("10_1", 2), None), ((10, 2), None)], key=lambda x: x[0])

    def test_unknown_types(self):
        orig = [
            LikeFloat(4),  # 0
            Comparable('hello'),  # 1
            LikeFloat(1),  # 2
            2.0,  # 3
            Comparable('world'),  # 4
            ToStr(1),  # 5
            NoStr('bogus'),  # 6
            ToStr('a'),  # 7
            ToStr('A'),  # 8
            3,  # 9
        ]

        ref = [orig[i] for i in (1, 4, 6, 5, 8, 7, 2, 3, 9, 0)]
        ans = sorted_robust(orig)
        self.assertEqual(len(orig), len(ans))
        for _r, _a in zip(ref, ans):
            self.assertIs(_r, _a)
        self.assertEqual(_robust_sort_keyfcn._typemap[LikeFloat], (1, float.__name__))
        self.assertEqual(
            _robust_sort_keyfcn._typemap[Comparable], (1, Comparable.__name__)
        )
        self.assertEqual(_robust_sort_keyfcn._typemap[ToStr], (2, ToStr.__name__))
        self.assertEqual(_robust_sort_keyfcn._typemap[NoStr], (3, NoStr.__name__))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Functions for solving the nonlinear subproblem."""

from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.common.errors import InfeasibleConstraintException, DeveloperError
from pyomo.contrib import appsi
from pyomo.contrib.appsi.cmodel import cmodel_available
from pyomo.contrib.fbbt.fbbt import fbbt
from pyomo.contrib.gdpopt.solve_discrete_problem import (
    distinguish_mip_infeasible_or_unbounded,
)
from pyomo.contrib.gdpopt.util import (
    SuppressInfeasibleWarning,
    is_feasible,
    get_main_elapsed_time,
)
from pyomo.core import Constraint, TransformationFactory, Objective, Block
import pyomo.core.expr as EXPR
from pyomo.opt import SolverFactory, SolverResults
from pyomo.opt import TerminationCondition as tc


def configure_and_call_solver(model, solver, args, problem_type, timing, time_limit):
    opt = SolverFactory(solver)
    if not opt.available():
        raise RuntimeError("%s solver %s is not available." % (problem_type, solver))
    with SuppressInfeasibleWarning():
        solver_args = dict(args)
        if time_limit is not None:
            elapsed = get_main_elapsed_time(timing)
            remaining = max(time_limit - elapsed, 1)
            if solver == 'gams':
                solver_args['add_options'] = solver_args.get('add_options', [])
                solver_args['add_options'].append('option reslim=%s;' % remaining)
            elif solver == 'multisolve':
                solver_args['time_limit'] = min(
                    solver_args.get('time_limit', float('inf')), remaining
                )
        try:
            results = opt.solve(model, **solver_args)
        except ValueError as err:
            if 'Cannot load a SolverResults object with bad status: error' in str(err):
                results = SolverResults()
                results.solver.termination_condition = tc.error
                results.solver.message = str(err)
            else:
                raise
    return results


def process_nonlinear_problem_results(results, model, problem_type, config):
    """Processes the results object returned from the nonlinear solver.
    Returns one of TerminationCondition.optimal (for locally optimal or
    globally optimal since people use this as a heuristic),
    TerminationCondition.feasible (we have a solution with no guarantees),
    TerminationCondition.noSolution (we have no solution, with no guarantees
    of infeasibility), or TerminationCondition.infeasible.
    """
    logger = config.logger
    term_cond = results.solver.termination_condition
    if any(
        term_cond == cond
        for cond in (tc.optimal, tc.locallyOptimal, tc.globallyOptimal)
    ):
        # Since we let people use local solvers and settle for the heuristic, we
        # just let all these by.
        return tc.optimal
    elif term_cond == tc.feasible:
        return tc.feasible
    elif term_cond == tc.infeasible:
        logger.debug('%s subproblem was infeasible.' % problem_type)
        return tc.infeasible
    elif term_cond == tc.maxIterations:
        logger.debug(
            '%s subproblem failed to converge within iteration limit.' % problem_type
        )
        if is_feasible(model, config):
            logger.debug(
                'NLP solution is still feasible. '
                'Using potentially suboptimal feasible solution.'
            )
            return tc.feasible
        return False
    elif term_cond == tc.internalSolverError:
        # Possible that IPOPT had a restoration failure
        logger.debug(
            "%s solver had an internal failure: %s"
            % (problem_type, results.solver.message)
        )
        return tc.noSolution
    elif term_cond == tc.other and "Too few degrees of freedom" in str(
        results.solver.message
    ):
        # Possible IPOPT degrees of freedom error
        logger.debug(
            "Perhaps the subproblem solver has too few degrees of freedom: %s"
            % results.solver.message
        )
        return tc.infeasible
    elif term_cond == tc.other:
        logger.debug(
            "%s solver had a termination condition of 'other': %s"
            % (problem_type, results.solver.message)
        )
        return tc.noSolution
    elif term_cond == tc.error:
        logger.debug(
            "%s solver had a termination condition of 'error': "
            "%s" % (problem_type, results.solver.message)
        )
        return tc.noSolution
    elif term_cond == tc.maxTimeLimit:
        logger.debug(
            "%s subproblem failed to converge within time limit." % problem_type
        )
        if is_feasible(model, config):
            config.logger.debug(
                '%s solution is still feasible. '
                'Using potentially suboptimal feasible solution.' % problem_type
            )
            return tc.feasible
        return tc.noSolution
    elif term_cond == tc.intermediateNonInteger:
        config.logger.debug(
            "%s solver could not find feasible integer"
            " solution: %s" % (problem_type, results.solver.message)
        )
        return tc.noSolution
    elif term_cond == tc.unbounded:
        config.logger.debug(
            "The NLP subproblem is unbounded, meaning that the GDP is unbounded."
        )
        return tc.unbounded
    else:
        # This isn't the user's fault, but we give up--we don't know what's
        # going on.
        raise DeveloperError(
            'GDPopt unable to handle %s subproblem termination '
            'condition of %s. Results: %s' % (problem_type, term_cond, results)
        )


def solve_linear_subproblem(subproblem, config, timing):
    results = configure_and_call_solver(
        subproblem,
        config.mip_solver,
        config.mip_solver_args,
        'MIP',
        timing,
        config.time_limit,
    )
    subprob_terminate_cond = results.solver.termination_condition
    if subprob_terminate_cond is tc.optimal:
        return tc.optimal
    elif subprob_terminate_cond is tc.infeasibleOrUnbounded:
        results, subprob_terminate_cond = distinguish_mip_infeasible_or_unbounded(
            subproblem, config
        )
    if subprob_terminate_cond is tc.infeasible:
        config.logger.debug('MILP subproblem was infeasible.')
        return tc.infeasible
    elif subprob_terminate_cond is tc.unbounded:
        config.logger.debug('MILP subproblem was unbounded.')
        return tc.unbounded
    else:
        raise ValueError(
            'GDPopt unable to handle MIP subproblem termination '
            'condition of %s. Results: %s' % (subprob_terminate_cond, results)
        )


def solve_NLP(nlp_model, config, timing):
    """Solve the NLP subproblem."""
    config.logger.debug(
        'Solving nonlinear subproblem for fixed binaries and logical realizations.'
    )

    results = configure_and_call_solver(
        nlp_model,
        config.nlp_solver,
        config.nlp_solver_args,
        'NLP',
        timing,
        config.time_limit,
    )

    return process_nonlinear_problem_results(results, nlp_model, 'NLP', config)


def solve_MINLP(util_block, config, timing):
    """Solve the MINLP subproblem."""
    config.logger.debug("Solving MINLP subproblem for fixed logical realizations.")
    model = util_block.parent_block()

    minlp_solver = SolverFactory(config.minlp_solver)
    if not minlp_solver.available():
        raise RuntimeError("MINLP solver %s is not available." % config.minlp_solver)

    results = configure_and_call_solver(
        model,
        config.minlp_solver,
        config.minlp_solver_args,
        'MINLP',
        timing,
        config.time_limit,
    )
    subprob_termination = process_nonlinear_problem_results(
        results, model, 'MINLP', config
    )

    return subprob_termination


def detect_unfixed_discrete_vars(model):
    """Detect unfixed discrete variables in use on the model."""
    var_set = ComponentSet()
    for constr in model.component_data_objects(
        Constraint, active=True, descend_into=True
    ):
        var_set.update(
            v
            for v in EXPR.identify_variables(constr.body, include_fixed=False)
            if not v.is_continuous()
        )
    for obj in model.component_data_objects(Objective, active=True):
        var_set.update(
            v
            for v in EXPR.identify_variables(obj.expr, include_fixed=False)
            if not v.is_continuous()
        )
    return var_set


class preprocess_subproblem:
    def __init__(self, util_block, config):
        self.util_block = util_block
        self.config = config

        self.not_infeas = True
        self.unfixed_vars = []
        self.original_bounds = ComponentMap()
        self.constraints_deactivated = []
        self.constraints_modified = {}

    def __enter__(self):
        """Applies preprocessing transformations to the model."""
        m = self.util_block.parent_block()

        # Save bounds so we can restore them
        for cons in m.component_data_objects(
            Constraint, active=True, descend_into=Block
        ):
            for v in EXPR.identify_variables(cons.expr):
                if v not in self.original_bounds.keys():
                    self.original_bounds[v] = (v.lb, v.ub)
                    if not v.fixed:
                        self.unfixed_vars.append(v)
        # We could miss if there is a variable that only appears in the
        # objective, but its bounds are not going to get changed anyway if
        # that's the case.

        try:
            # First do FBBT
            # When #2574 is resolved, we can do the below. For now
            # we'll use contrib.fbbt
            # if cmodel_available:
            #     # use the appsi fbbt implementation since we can
            #     it = appsi.fbbt.IntervalTightener()
            #     it.config.integer_tol = self.config.integer_tolerance
            #     it.config.feasibility_tol = self.config.constraint_tolerance
            #     it.config.max_iter = self.config.max_fbbt_iterations
            #     it.perform_fbbt(m)
            fbbt(
                m,
                integer_tol=self.config.integer_tolerance,
                feasibility_tol=self.config.constraint_tolerance,
                max_iter=self.config.max_fbbt_iterations,
            )
            xfrm = TransformationFactory
            # Now that we've tightened bounds, see if any variables are fixed
            # because their lb is equal to the ub (within tolerance)
            xfrm('contrib.detect_fixed_vars').apply_to(
                m, tolerance=self.config.variable_tolerance
            )

            # Restore the original bounds because the subproblem solver might
            # like that better and because, if deactivate_trivial_constraints
            # ever gets fancier, this could change what is and is not trivial.
            if not self.config.tighten_nlp_var_bounds:
                for v, (lb, ub) in self.original_bounds.items():
                    v.setlb(lb)
                    v.setub(ub)

            # Now, if something got fixed to 0, we might have 0*var terms to
            # remove
            xfrm('contrib.remove_zero_terms').apply_to(
                m, constraints_modified=self.constraints_modified
            )
            # Last, check if any constraints are now trivial and deactivate them
            xfrm('contrib.deactivate_trivial_constraints').apply_to(
                m,
                tolerance=self.config.constraint_tolerance,
                return_trivial=self.constraints_deactivated,
            )

        except InfeasibleConstraintException as e:
            self.config.logger.debug(
                "NLP subproblem determined to be infeasible "
                "during preprocessing. Message: %s" % e
            )
            self.not_infeas = False

        return self.not_infeas

    def __exit__(self, type, value, traceback):
        # restore the bounds if we found the problem infeasible or if we didn't
        # do it above
        if not self.not_infeas or self.config.tighten_nlp_var_bounds:
            for v, (lb, ub) in self.original_bounds.items():
                v.setlb(lb)
                v.setub(ub)

        # A bit counter-intuitively (but I assume so that it can propagate those
        # bounds elsewhere), fbbt tightens the bounds on the fixed Boolean vars,
        # so we restore the bounds here
        for disj in self.util_block.disjunct_list:
            disj.binary_indicator_var.setlb(0)
            disj.binary_indicator_var.setub(1)
        for bool_var in self.util_block.non_indicator_boolean_variable_list:
            bool_var.get_associated_binary().setlb(0)
            bool_var.get_associated_binary().setub(1)

        # reactivate constraints
        for cons in self.constraints_deactivated:
            cons.activate()

        for cons, (orig, modified) in self.constraints_modified.items():
            cons.set_value(orig)

        # unfix variables:
        for v in self.unfixed_vars:
            v.unfix()


def call_appropriate_subproblem_solver(subprob_util_block, solver, config):
    timing = solver.timing
    subprob = subprob_util_block.parent_block()
    config.call_before_subproblem_solve(solver, subprob, subprob_util_block)

    # Is the subproblem linear?
    if not any(
        constr.body.polynomial_degree() not in (1, 0)
        for constr in subprob.component_data_objects(Constraint, active=True)
    ):
        subprob_termination = solve_linear_subproblem(subprob, config, timing)
    else:
        # Does it have any discrete variables, and is that allowed?
        unfixed_discrete_vars = detect_unfixed_discrete_vars(subprob)
        if config.force_subproblem_nlp and len(unfixed_discrete_vars) > 0:
            # this is actually our fault at this point--we should have
            # enumerated the discrete solutions if it was possible and the user
            # requested.
            raise DeveloperError(
                "Unfixed discrete variables found on the NLP subproblem."
            )
        elif len(unfixed_discrete_vars) == 0:
            subprob_termination = solve_NLP(subprob, config, timing)
        else:
            config.logger.debug(
                "The following discrete variables are unfixed: %s"
                "\nProceeding by solving the subproblem as a MINLP."
                % ", ".join([v.name for v in unfixed_discrete_vars])
            )
            subprob_termination = solve_MINLP(subprob_util_block, config, timing)

    # Call the NLP post-solve callback
    config.call_after_subproblem_solve(solver, subprob, subprob_util_block)

    # if feasible, call the NLP post-feasible callback
    if subprob_termination in {tc.optimal, tc.feasible}:
        config.call_after_subproblem_feasible(solver, subprob, subprob_util_block)

    return subprob_termination


def solve_subproblem(subprob_util_block, solver, config):
    """Set up and solve the local MINLP or NLP subproblem."""
    if config.subproblem_presolve:
        with preprocess_subproblem(subprob_util_block, config) as call_solver:
            if call_solver:
                return call_appropriate_subproblem_solver(
                    subprob_util_block, solver, config
                )
            else:
                return tc.infeasible

    return call_appropriate_subproblem_solver(subprob_util_block, solver, config)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.collections import Bunch
from pyomo.dataportal.process_data import _process_data


class TableData:
    """
    A class used to read/write data from/to a table in an external
    data source.
    """

    def __init__(self):
        """
        Constructor
        """
        self._info = None
        self._data = None
        self.options = Bunch()
        self.options.ncolumns = 1

    def available(self):
        """
        Returns:
            Return :const:`True` if the data manager is available.
        """
        return True

    def initialize(self, **kwds):
        """
        Initialize the data manager with keyword arguments.

        The `filename` argument is recognized here, and other arguments
        are passed to the :func:`add_options` method.
        """
        self.filename = kwds.pop('filename')
        self.add_options(**kwds)

    def add_options(self, **kwds):
        """
        Add the keyword options to the :class:`Options` object in this
        object.
        """
        self.options.update(kwds)

    def open(self):  # pragma:nocover
        """
        Open the data manager.
        """
        pass

    def read(self):  # pragma:nocover
        """
        Read data from the data manager.
        """
        return False

    def write(self, data):  # pragma:nocover
        """
        Write data to the data manager.
        """
        return False

    def close(self):  # pragma:nocover
        """
        Close the data manager.
        """
        pass

    def process(self, model, data, default):
        """
        Process the data that was extracted from this data manager and
        return it.
        """
        if model is None:
            model = self.options.model
        if not self.options.namespace in data:
            data[self.options.namespace] = {}
        return _process_data(
            self._info,
            model,
            data[self.options.namespace],
            default,
            self.filename,
            index=self.options.index,
            set=self.options.set,
            param=self.options.param,
            ncolumns=self.options.ncolumns,
        )

    def clear(self):
        """
        Clear the data that was extracted from this table
        """
        self._info = None

    def _set_data(self, headers, rows):
        from pyomo.core.base.set import Set
        from pyomo.core.base.param import Param

        header_index = []
        if self.options.select is None:
            for i in range(len(headers)):
                header_index.append(i)
        else:
            for i in self.options.select:
                try:
                    header_index.append(headers.index(str(i)))
                except:
                    print(
                        "Model declaration '%s' not found in returned query columns"
                        % str(i)
                    )
                    raise
        self.options.ncolumns = len(headers)

        if not self.options.param is None:
            if not type(self.options.param) in (list, tuple):
                self.options.param = (self.options.param,)
            _params = []
            for p in self.options.param:
                if isinstance(p, Param):
                    self.options.model = p.model()
                    _params.append(p.local_name)
                else:
                    _params.append(p)
            self.options.param = tuple(_params)

        if isinstance(self.options.set, Set):
            self.options.model = self.options.set.model()
            self.options.set = self.options.set.local_name

        if isinstance(self.options.index, Set):
            self.options.model = self.options.index.model()
            self.options.index = self.options.index.local_name

        elif type(self.options.index) in [tuple, list]:
            tmp = []
            for val in self.options.index:
                if isinstance(val, Set):
                    tmp.append(val.local_name)
                    self.options.model = val.model()
                else:
                    tmp.append(val)
            self.options.index = tuple(tmp)

        if self.options.format is None:
            if not self.options.set is None:
                self.options.format = 'set'
            elif not self.options.param is None:
                self.options.format = 'table'
            if self.options.format is None:
                raise ValueError("Unspecified format and  data option")
        elif self.options.set is None and self.options.param is None:
            msg = "Must specify the set or parameter option for data"
            raise IOError(msg)

        if self.options.format == 'set':
            if not self.options.index is None:
                msg = "Cannot specify index for data with the 'set' format: %s"
                raise IOError(msg % str(self.options.index))

            self._info = ["set", self.options.set, ":="]
            for row in rows:
                if self.options.ncolumns > 1:
                    self._info.append(tuple(row))
                else:
                    self._info.extend(row)

        elif self.options.format == 'set_array':
            if not self.options.index is None:
                msg = "Cannot specify index for data with the 'set_array' format: %s"
                raise IOError(msg % str(self.options.index))

            self._info = ["set", self.options.set, ":"]
            self._info.extend(headers[1:])
            self._info.append(":=")
            for row in rows:
                self._info.extend(row)

        elif self.options.format == 'transposed_array':
            self._info = ["param", self.options.param[0], "(tr)", ":"]
            self._info.extend(headers[1:])
            self._info.append(":=")
            for row in rows:
                self._info.extend(row)

        elif self.options.format == 'array':
            self._info = ["param", self.options.param[0], ":"]
            self._info.extend(headers[1:])
            self._info.append(":=")
            for row in rows:
                self._info.extend(row)

        elif self.options.format == 'table':
            if self.options.index is not None:
                self._info = ["param", ":", self.options.index, ":"]
            else:
                self._info = ["param", ":"]
            for param in self.options.param:
                self._info.append(param)
            self._info.append(":=")
            for row in rows:
                for i in header_index:
                    self._info.append(row[i])
            self.options.ncolumns = len(header_index)
        else:
            msg = "Unknown parameter format: '%s'"
            raise ValueError(msg % self.options.format)

    def _get_table(self):
        from pyomo.core.expr import value

        tmp = []
        if self.options.columns is not None:
            tmp.append(self.options.columns)
        if self.options.set is not None:
            # Create column names
            if self.options.columns is None:
                cols = []
                for i in range(self.options.set.dimen):
                    cols.append(self.options.set.local_name + str(i))
                tmp.append(cols)
            # Get rows
            if self.options.sort is not None:
                for data in sorted(self.options.set):
                    if self.options.set.dimen > 1:
                        tmp.append(list(data))
                    else:
                        tmp.append([data])
            else:
                for data in self.options.set:
                    if self.options.set.dimen > 1:
                        tmp.append(list(data))
                    else:
                        tmp.append([data])
        elif self.options.param is not None:
            if type(self.options.param) in (list, tuple):
                _param = self.options.param
            else:
                _param = [self.options.param]
            # Collect data
            for index in _param[0]:
                if index is None:
                    row = []
                elif type(index) in (list, tuple):
                    row = list(index)
                else:
                    row = [index]
                for param in _param:
                    row.append(value(param[index]))
                tmp.append(row)
            # Create column names
            if self.options.columns is None:
                cols = []
                for i in range(len(tmp[0]) - len(_param)):
                    cols.append('I' + str(i))
                for param in _param:
                    cols.append(param)
                tmp.insert(0, cols)
        return tmp



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import collections.abc
import pickle

import pyomo.common.unittest as unittest
import pyomo.kernel as pmo
from pyomo.core.kernel.base import ICategorizedObject, ICategorizedObjectContainer
from pyomo.core.kernel.homogeneous_container import IHomogeneousContainer
from pyomo.core.kernel.tuple_container import TupleContainer
from pyomo.core.kernel.block import block, block_list

#
# There are no fully implemented test suites in this
# file. These classes are meant to be used to test
# full implementations of TupleContainer containers.
# To test a TupleContainer implementation, just import
# this class and use it as a subclass in another test suite.
#

# Note: we need to test with a pickle protocol
#       that is high enough to support __slots__
#       and weakref (bas
_pickle_test_protocol = pickle.HIGHEST_PROTOCOL


class _bad_ctype:
    ctype = "_this_is_definitely_not_the_ctype_being_tested"


class _TestTupleContainerBase:
    # set by derived class
    _container_type = None
    _ctype_factory = None

    def test_ctype(self):
        c = self._container_type()
        ctype = self._ctype_factory().ctype
        self.assertIs(c.ctype, ctype)
        self.assertIs(type(c)._ctype, ctype)
        self.assertIs(self._container_type._ctype, ctype)

    def test_init1(self):
        ctuple = self._container_type()

    def test_init2(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        with self.assertRaises(TypeError):
            d = self._container_type(*tuple(self._ctype_factory() for i in index))

    def test_type(self):
        ctuple = self._container_type()
        self.assertTrue(isinstance(ctuple, ICategorizedObject))
        self.assertTrue(isinstance(ctuple, ICategorizedObjectContainer))
        self.assertTrue(isinstance(ctuple, IHomogeneousContainer))
        self.assertTrue(isinstance(ctuple, TupleContainer))
        self.assertTrue(isinstance(ctuple, collections.abc.Sequence))
        self.assertTrue(issubclass(type(ctuple), collections.abc.Sequence))

    def test_len1(self):
        c = self._container_type()
        self.assertEqual(len(c), 0)

    def test_len2(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        self.assertEqual(len(c), len(index))

    def test_wrong_type_init(self):
        index = range(5)
        with self.assertRaises(TypeError):
            c = self._container_type(_bad_ctype() for i in index)

    def test_has_parent_init(self):
        ctuple = self._container_type([self._ctype_factory()])
        with self.assertRaises(ValueError):
            self._container_type([ctuple[0]])

    def test_iter(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        self.assertEqual(len(c), len(index))
        raw_tuple = c[:]
        self.assertEqual(type(raw_tuple), tuple)
        for c1, c2 in zip(raw_tuple, c):
            self.assertEqual(id(c1), id(c2))

    def test_reverse(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        raw_tuple = c[:]
        self.assertEqual(type(raw_tuple), tuple)
        for c1, c2 in zip(reversed(c), reversed(raw_tuple)):
            self.assertEqual(id(c1), id(c2))

    def test_index(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        for i in index:
            cdata = c[i]
            self.assertEqual(c.index(cdata), i)
            self.assertEqual(c.index(cdata, start=i), i)
            with self.assertRaises(ValueError):
                c.index(cdata, start=i + 1)
            with self.assertRaises(ValueError):
                c.index(cdata, start=i, stop=i)
            with self.assertRaises(ValueError):
                c.index(cdata, stop=i)
            self.assertEqual(c.index(cdata, start=i, stop=i + 1), i)
            with self.assertRaises(ValueError):
                c.index(cdata, start=i + 1, stop=i + 1)
            self.assertEqual(c.index(cdata, start=-len(index) + i), i)
            if i == index[-1]:
                self.assertEqual(c.index(cdata, start=-len(index) + i + 1), i)
            else:
                with self.assertRaises(ValueError):
                    self.assertEqual(c.index(cdata, start=-len(index) + i + 1), i)
            if i == index[-1]:
                with self.assertRaises(ValueError):
                    self.assertEqual(c.index(cdata, stop=-len(index) + i + 1), i)
            else:
                self.assertEqual(c.index(cdata, stop=-len(index) + i + 1), i)
        tmp = self._ctype_factory()
        with self.assertRaises(ValueError):
            c.index(tmp)
        with self.assertRaises(ValueError):
            c.index(tmp, stop=len(c) + 1)

    def test_count(self):
        index = range(5)
        c = self._container_type(self._ctype_factory() for i in index)
        for i in index:
            self.assertEqual(c.count(c[i]), 1)

    def test_pickle(self):
        index = range(5)
        ctuple = self._container_type(
            [self._ctype_factory() for i in index] + [self._container_type()]
        )
        index = list(index)
        index = index + [len(index)]
        for i in index:
            self.assertTrue(ctuple[i].parent is ctuple)
        pickled_ctuple = pickle.loads(
            pickle.dumps(ctuple, protocol=_pickle_test_protocol)
        )
        self.assertTrue(isinstance(pickled_ctuple, self._container_type))
        self.assertTrue(pickled_ctuple.parent is None)
        self.assertEqual(len(pickled_ctuple), len(index))
        self.assertNotEqual(id(pickled_ctuple), id(ctuple))
        for i in index:
            self.assertNotEqual(id(pickled_ctuple[i]), id(ctuple[i]))
            self.assertTrue(pickled_ctuple[i].parent is pickled_ctuple)
            self.assertTrue(ctuple[i].parent is ctuple)

    def test_eq(self):
        ctuple1 = self._container_type([self._ctype_factory()])
        ctuple2 = self._container_type([self._ctype_factory()])

        self.assertNotEqual(ctuple1, set())
        self.assertFalse(ctuple1 == set())
        self.assertNotEqual(ctuple1, list())
        self.assertFalse(ctuple1 == list())
        self.assertNotEqual(ctuple1, tuple())
        self.assertFalse(ctuple1 == tuple())
        self.assertNotEqual(ctuple1, dict())
        self.assertFalse(ctuple1 == dict())

        self.assertTrue(ctuple1 == ctuple1)
        self.assertEqual(ctuple1, ctuple1)
        self.assertTrue(ctuple1 == list(ctuple1))
        self.assertEqual(ctuple1, list(ctuple1))
        self.assertTrue(ctuple1 == tuple(ctuple1))
        self.assertEqual(ctuple1, tuple(ctuple1))
        self.assertTrue(list(ctuple1) == ctuple1)
        self.assertEqual(list(ctuple1), ctuple1)
        self.assertTrue(tuple(ctuple1) == ctuple1)
        self.assertEqual(tuple(ctuple1), ctuple1)

        self.assertFalse(ctuple2 == ctuple1)
        self.assertTrue(ctuple2 != ctuple1)
        self.assertNotEqual(ctuple2, ctuple1)

        self.assertFalse(ctuple1 == ctuple2)
        self.assertTrue(ctuple1 != ctuple2)
        self.assertNotEqual(ctuple1, ctuple2)

        ctuple1 = self._container_type()

        self.assertEqual(ctuple1, set())
        self.assertTrue(ctuple1 == set())
        self.assertEqual(ctuple1, list())
        self.assertTrue(ctuple1 == list())
        self.assertEqual(ctuple1, tuple())
        self.assertTrue(ctuple1 == tuple())
        self.assertNotEqual(ctuple1, dict())
        self.assertFalse(ctuple1 == dict())

    def test_child(self):
        ctuple = self._container_type()
        with self.assertRaises(KeyError):
            ctuple.child(0)
        c = self._ctype_factory()
        ctuple = self._container_type([c])
        self.assertIs(ctuple.child(0), c)

    def test_name(self):
        children = [self._ctype_factory() for i in range(5)]
        children.append(self._container_type([self._ctype_factory()]))

        for c in children:
            self.assertTrue(c.parent is None)
            self.assertEqual(c.local_name, None)
            self.assertEqual(c.name, None)

        ctuple = self._container_type(children)
        self.assertTrue(ctuple.parent is None)
        self.assertEqual(ctuple.local_name, None)
        self.assertEqual(ctuple.name, None)
        names = pmo.generate_names(ctuple)
        for i, c in enumerate(children):
            self.assertTrue(c.parent is ctuple)
            self.assertEqual(c.local_name, "[%s]" % (i))
            self.assertEqual(c.name, "[%s]" % (i))
            self.assertEqual(c.name, names[c])
        for c in ctuple.components():
            self.assertNotEqual(c.name, None)
            self.assertEqual(c.name, names[c])

        model = block()
        model.ctuple = ctuple
        self.assertTrue(model.parent is None)
        self.assertTrue(ctuple.parent is model)
        self.assertEqual(ctuple.local_name, "ctuple")
        self.assertEqual(ctuple.name, "ctuple")
        names = pmo.generate_names(model)
        for i, c in enumerate(children):
            self.assertTrue(c.parent is ctuple)
            self.assertEqual(c.local_name, "[%s]" % (i))
            self.assertEqual(c.name, "ctuple[%s]" % (i))
            self.assertEqual(c.name, names[c])
        for c in ctuple.components():
            self.assertNotEqual(c.name, None)
            self.assertEqual(c.name, names[c])

        b = block()
        b.model = model
        self.assertTrue(b.parent is None)
        self.assertTrue(model.parent is b)
        self.assertTrue(ctuple.parent is model)
        self.assertEqual(ctuple.local_name, "ctuple")
        self.assertEqual(ctuple.name, "model.ctuple")
        names = pmo.generate_names(b)
        for i, c in enumerate(children):
            self.assertTrue(c.parent is ctuple)
            self.assertEqual(c.local_name, "[%s]" % (i))
            self.assertEqual(c.name, "model.ctuple[%s]" % (i))
            self.assertEqual(c.name, names[c])
        for c in ctuple.components():
            self.assertNotEqual(c.name, None)
            self.assertEqual(c.name, names[c])

        blist = block_list()
        blist.append(b)
        self.assertTrue(blist.parent is None)
        self.assertTrue(b.parent is blist)
        self.assertTrue(model.parent is b)
        self.assertTrue(ctuple.parent is model)
        self.assertEqual(ctuple.local_name, "ctuple")
        self.assertEqual(ctuple.name, "[0].model.ctuple")
        for i, c in enumerate(children):
            self.assertTrue(c.parent is ctuple)
            self.assertEqual(c.local_name, "[%s]" % (i))
            self.assertEqual(c.name, "[0].model.ctuple[%s]" % (i))

        m = block()
        m.blist = blist
        self.assertTrue(m.parent is None)
        self.assertTrue(blist.parent is m)
        self.assertTrue(b.parent is blist)
        self.assertTrue(model.parent is b)
        self.assertTrue(ctuple.parent is model)
        self.assertEqual(ctuple.local_name, "ctuple")
        self.assertEqual(ctuple.name, "blist[0].model.ctuple")
        names = pmo.generate_names(m)
        for i, c in enumerate(children):
            self.assertTrue(c.parent is ctuple)
            self.assertEqual(c.local_name, "[%s]" % (i))
            self.assertEqual(c.name, "blist[0].model.ctuple[%s]" % (i))
            self.assertEqual(c.name, names[c])
        for c in ctuple.components():
            self.assertNotEqual(c.name, None)
            self.assertEqual(c.name, names[c])
        names = pmo.generate_names(m)
        for c in m.children():
            self.assertEqual(c.name, names[c])

    def test_components(self):
        ctuple = self._container_type()
        self.assertEqual(list(ctuple.components()), [])

        ctupleflattened = []
        ctupleflattened.append(self._ctype_factory())
        ctupleflattened.append(self._ctype_factory())

        csubtupleflattened = []
        csubtupleflattened.append(self._ctype_factory())
        ctupleflattened.append(csubtupleflattened[-1])

        csubtupleflattened.append(self._ctype_factory())
        ctupleflattened.append(csubtupleflattened[-1])

        csubtupleflattened.append(self._ctype_factory())
        ctupleflattened.append(csubtupleflattened[-1])

        csubtuple = self._container_type(csubtupleflattened)
        self.assertEqual(
            list(id(_c) for _c in csubtuple.components()),
            list(id(_c) for _c in csubtupleflattened),
        )
        self.assertEqual(
            len(set(id(_c) for _c in csubtuple.components())),
            len(list(id(_c) for _c in csubtuple.components())),
        )
        self.assertEqual(len(set(id(_c) for _c in csubtuple.components())), 3)

        ctuple = self._container_type(
            [ctupleflattened[0], ctupleflattened[1], csubtuple]
        )
        self.assertEqual(
            list(id(_c) for _c in ctuple.components()),
            list(id(_c) for _c in ctupleflattened),
        )
        self.assertEqual(
            len(set(id(_c) for _c in ctuple.components())),
            len(list(id(_c) for _c in ctuple.components())),
        )
        self.assertEqual(len(set(id(_c) for _c in ctuple.components())), 5)

    def test_preorder_traversal(self):
        csubtuple = self._container_type([self._ctype_factory()])
        ctuple = self._container_type(
            [self._ctype_factory(), csubtuple, self._ctype_factory()]
        )

        traversal = []
        traversal.append(ctuple)
        traversal.append(ctuple[0])
        traversal.append(ctuple[1])
        traversal.append(ctuple[1][0])
        traversal.append(ctuple[2])

        self.assertEqual(
            [c.name for c in traversal],
            [c.name for c in pmo.preorder_traversal(ctuple)],
        )
        self.assertEqual(
            [id(c) for c in traversal], [id(c) for c in pmo.preorder_traversal(ctuple)]
        )

        return ctuple, traversal

    def test_preorder_traversal_descend_check(self):
        csubtuple = self._container_type([self._ctype_factory()])
        ctuple = self._container_type(
            [self._ctype_factory(), csubtuple, self._ctype_factory()]
        )

        traversal = []
        traversal.append(ctuple)
        traversal.append(ctuple[0])
        traversal.append(ctuple[1])
        traversal.append(ctuple[1][0])
        traversal.append(ctuple[2])

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return False

        descend.seen = []
        order = list(pmo.preorder_traversal(ctuple, descend=descend))
        self.assertEqual(len(order), 1)
        self.assertIs(order[0], ctuple)
        self.assertEqual(len(descend.seen), 1)
        self.assertIs(descend.seen[0], ctuple)

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return True

        descend.seen = []
        order = list(pmo.preorder_traversal(ctuple, descend=descend))
        self.assertEqual([c.name for c in traversal], [c.name for c in order])
        self.assertEqual([id(c) for c in traversal], [id(c) for c in order])
        self.assertEqual(
            [c.name for c in traversal if c._is_container],
            [c.name for c in descend.seen],
        )
        self.assertEqual(
            [id(c) for c in traversal if c._is_container], [id(c) for c in descend.seen]
        )

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return True

        descend.seen = []
        order = list(pmo.preorder_traversal(ctuple, descend=descend))
        self.assertEqual([c.name for c in traversal], [c.name for c in order])
        self.assertEqual([id(c) for c in traversal], [id(c) for c in order])
        self.assertEqual(
            [c.name for c in traversal if c._is_container],
            [c.name for c in descend.seen],
        )
        self.assertEqual(
            [id(c) for c in traversal if c._is_container], [id(c) for c in descend.seen]
        )
        return ctuple, traversal


class _TestActiveTupleContainerBase(_TestTupleContainerBase):
    def test_active_type(self):
        ctuple = self._container_type()
        self.assertTrue(isinstance(ctuple, ICategorizedObject))
        self.assertTrue(isinstance(ctuple, ICategorizedObjectContainer))
        self.assertTrue(isinstance(ctuple, IHomogeneousContainer))
        self.assertTrue(isinstance(ctuple, TupleContainer))
        self.assertTrue(isinstance(ctuple, collections.abc.Sequence))
        self.assertTrue(issubclass(type(ctuple), collections.abc.Sequence))

    def test_active(self):
        index = list(range(4))
        ctuple = self._container_type(self._ctype_factory() for i in index)
        with self.assertRaises(AttributeError):
            ctuple.active = False
        for c in ctuple:
            with self.assertRaises(AttributeError):
                c.active = False

        model = block()
        model.ctuple = ctuple
        b = block()
        b.model = model
        blist = block_list()
        blist.append(b)
        blist.append(block())
        m = block()
        m.blist = blist

        self.assertEqual(m.active, True)
        self.assertEqual(blist.active, True)
        self.assertEqual(blist[1].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(ctuple.active, True)
        for c in ctuple:
            self.assertEqual(c.active, True)
        for c in ctuple.components():
            self.assertEqual(c.active, True)
        for c in ctuple.components(active=True):
            self.assertEqual(c.active, True)
        self.assertEqual(len(list(ctuple.components())), len(ctuple))
        self.assertEqual(
            len(list(ctuple.components())), len(list(ctuple.components(active=True)))
        )

        m.deactivate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(ctuple.active, False)
        for c in ctuple:
            self.assertEqual(c.active, False)
        self.assertNotEqual(
            len(list(ctuple.components())), len(list(ctuple.components(active=None)))
        )
        self.assertEqual(len(list(ctuple.components(active=True))), 0)

        test_c = ctuple[0]
        test_c.activate()
        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(ctuple.active, False)
        ctuple.activate()
        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(ctuple.active, True)
        for c in ctuple:
            if c is test_c:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for c in ctuple.components():
            if c is test_c:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for c in ctuple.components(active=True):
            self.assertEqual(c.active, True)
        self.assertNotEqual(
            len(list(ctuple.components())), len(list(ctuple.components(active=None)))
        )
        self.assertEqual(len(list(ctuple.components(active=True))), 1)

        m.activate(shallow=False)

        self.assertEqual(m.active, True)
        self.assertEqual(blist.active, True)
        self.assertEqual(blist[1].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(ctuple.active, True)
        for c in ctuple:
            self.assertEqual(c.active, True)
        for c in ctuple.components():
            self.assertEqual(c.active, True)
        for c in ctuple.components(active=True):
            self.assertEqual(c.active, True)
        self.assertEqual(len(list(ctuple.components())), len(ctuple))
        self.assertEqual(
            len(list(ctuple.components())), len(list(ctuple.components(active=True)))
        )

        m.deactivate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(ctuple.active, False)
        for c in ctuple:
            self.assertEqual(c.active, False)
        self.assertNotEqual(
            len(list(ctuple.components())), len(list(ctuple.components(active=None)))
        )
        self.assertEqual(len(list(ctuple.components(active=True))), 0)

        ctuple.activate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(ctuple.active, True)
        for i, c in enumerate(ctuple):
            self.assertEqual(c.active, True)
        for c in ctuple.components():
            self.assertEqual(c.active, True)
        for c in ctuple.components(active=True):
            self.assertEqual(c.active, True)
        self.assertEqual(len(list(ctuple.components())), len(ctuple))
        self.assertEqual(
            len(list(ctuple.components())), len(list(ctuple.components(active=True)))
        )

        ctuple.deactivate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(ctuple.active, False)
        for i, c in enumerate(ctuple):
            self.assertEqual(c.active, False)
        self.assertNotEqual(
            len(list(ctuple.components())), len(list(ctuple.components(active=None)))
        )
        self.assertEqual(len(list(ctuple.components(active=True))), 0)

        ctuple[-1].activate()

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(ctuple.active, False)
        ctuple.activate()
        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(ctuple.active, True)
        for i, c in enumerate(ctuple):
            if i == len(ctuple) - 1:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for i, c in enumerate(ctuple.components(active=None)):
            if i == len(ctuple) - 1:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for c in ctuple.components():
            self.assertEqual(c.active, True)
        for c in ctuple.components(active=True):
            self.assertEqual(c.active, True)
        self.assertNotEqual(
            len(list(ctuple.components())), len(list(ctuple.components(active=None)))
        )
        self.assertEqual(len(list(ctuple.components(active=True))), 1)

        ctuple.deactivate(shallow=False)
        ctuple.activate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(blist.active, False)
        self.assertEqual(blist[1].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(ctuple.active, True)
        for i, c in enumerate(ctuple):
            self.assertEqual(c.active, True)
        for c in ctuple.components():
            self.assertEqual(c.active, True)
        for c in ctuple.components(active=True):
            self.assertEqual(c.active, True)
        self.assertEqual(len(list(ctuple.components())), len(ctuple))
        self.assertEqual(
            len(list(ctuple.components())), len(list(ctuple.components(active=True)))
        )

    def test_preorder_traversal(self):
        ctuple, traversal = super(
            _TestActiveTupleContainerBase, self
        ).test_preorder_traversal()

        ctuple[1].deactivate()
        self.assertEqual(
            [None, '[0]', '[2]'],
            [c.name for c in pmo.preorder_traversal(ctuple, active=True)],
        )
        self.assertEqual(
            [id(ctuple), id(ctuple[0]), id(ctuple[2])],
            [id(c) for c in pmo.preorder_traversal(ctuple, active=True)],
        )

        ctuple[1].deactivate(shallow=False)
        self.assertEqual(
            [c.name for c in traversal if c.active],
            [c.name for c in pmo.preorder_traversal(ctuple, active=True)],
        )
        self.assertEqual(
            [id(c) for c in traversal if c.active],
            [id(c) for c in pmo.preorder_traversal(ctuple, active=True)],
        )

        ctuple.deactivate()
        self.assertEqual(len(list(pmo.preorder_traversal(ctuple, active=True))), 0)
        self.assertEqual(len(list(pmo.generate_names(ctuple, active=True))), 0)

    def test_preorder_traversal_descend_check(self):
        ctuple, traversal = super(
            _TestActiveTupleContainerBase, self
        ).test_preorder_traversal_descend_check()

        ctuple[1].deactivate()

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return True

        descend.seen = []
        order = list(pmo.preorder_traversal(ctuple, active=True, descend=descend))
        self.assertEqual([None, '[0]', '[2]'], [c.name for c in order])
        self.assertEqual(
            [id(ctuple), id(ctuple[0]), id(ctuple[2])], [id(c) for c in order]
        )
        if ctuple.ctype._is_heterogeneous_container:
            self.assertEqual([None, '[0]', '[2]'], [c.name for c in descend.seen])
            self.assertEqual(
                [id(ctuple), id(ctuple[0]), id(ctuple[2])],
                [id(c) for c in descend.seen],
            )
        else:
            self.assertEqual([None], [c.name for c in descend.seen])
            self.assertEqual([id(ctuple)], [id(c) for c in descend.seen])

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return x.active

        descend.seen = []
        order = list(pmo.preorder_traversal(ctuple, active=None, descend=descend))
        self.assertEqual([None, '[0]', '[1]', '[2]'], [c.name for c in order])
        self.assertEqual(
            [id(ctuple), id(ctuple[0]), id(ctuple[1]), id(ctuple[2])],
            [id(c) for c in order],
        )
        if ctuple.ctype._is_heterogeneous_container:
            self.assertEqual(
                [None, '[0]', '[1]', '[2]'], [c.name for c in descend.seen]
            )
            self.assertEqual(
                [id(ctuple), id(ctuple[0]), id(ctuple[1]), id(ctuple[2])],
                [id(c) for c in descend.seen],
            )
        else:
            self.assertEqual([None, '[1]'], [c.name for c in descend.seen])
            self.assertEqual([id(ctuple), id(ctuple[1])], [id(c) for c in descend.seen])

        ctuple[1].deactivate(shallow=False)

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return True

        descend.seen = []
        order = list(pmo.preorder_traversal(ctuple, active=True, descend=descend))
        self.assertEqual(
            [c.name for c in traversal if c.active], [c.name for c in order]
        )
        self.assertEqual([id(c) for c in traversal if c.active], [id(c) for c in order])
        self.assertEqual(
            [c.name for c in traversal if c.active and c._is_container],
            [c.name for c in descend.seen],
        )
        self.assertEqual(
            [id(c) for c in traversal if c.active and c._is_container],
            [id(c) for c in descend.seen],
        )

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return x.active

        descend.seen = []
        order = list(pmo.preorder_traversal(ctuple, active=None, descend=descend))
        self.assertEqual([None, '[0]', '[1]', '[2]'], [c.name for c in order])
        self.assertEqual(
            [id(ctuple), id(ctuple[0]), id(ctuple[1]), id(ctuple[2])],
            [id(c) for c in order],
        )
        if ctuple.ctype._is_heterogeneous_container:
            self.assertEqual(
                [None, '[0]', '[1]', '[2]'], [c.name for c in descend.seen]
            )
            self.assertEqual(
                [id(ctuple), id(ctuple[0]), id(ctuple[1]), id(ctuple[2])],
                [id(c) for c in descend.seen],
            )
        else:
            self.assertEqual([None, '[1]'], [c.name for c in descend.seen])
            self.assertEqual([id(ctuple), id(ctuple[1])], [id(c) for c in descend.seen])

        ctuple.deactivate()

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return True

        descend.seen = []
        order = list(pmo.preorder_traversal(ctuple, active=True, descend=descend))
        self.assertEqual(len(descend.seen), 0)
        self.assertEqual(len(list(pmo.generate_names(ctuple, active=True))), 0)

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return x.active

        descend.seen = []
        order = list(pmo.preorder_traversal(ctuple, active=None, descend=descend))
        self.assertEqual(len(descend.seen), 1)
        self.assertIs(descend.seen[0], ctuple)

        ctuple.deactivate(shallow=False)

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return True

        descend.seen = []
        order = list(pmo.preorder_traversal(ctuple, active=True, descend=descend))
        self.assertEqual(len(descend.seen), 0)
        self.assertEqual(len(list(pmo.generate_names(ctuple, active=True))), 0)

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return x.active

        descend.seen = []
        order = list(pmo.preorder_traversal(ctuple, active=None, descend=descend))
        self.assertEqual(len(descend.seen), 1)
        self.assertIs(descend.seen[0], ctuple)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.core.base import (
    Transformation,
    TransformationFactory,
    Constraint,
    Block,
    Param,
    SortComponents,
    ComponentUID,
)
from pyomo.mpec.complementarity import Complementarity
from pyomo.gdp import Disjunct

logger = logging.getLogger('pyomo.core')


#
# This transformation reworks each Complementarity block to
# add a constraint that ensures the complementarity condition.
# Specifically,
#
#   x1 >= 0  OR  x2 >= 0
#
# becomes
#
#   x1 >= 0
#   x2 >= 0
#   x1*x2 <= 0
#
@TransformationFactory.register(
    'mpec.simple_nonlinear',
    doc="Nonlinear transformations of complementarity conditions when all variables are non-negative",
)
class MPEC1_Transformation(Transformation):
    def __init__(self):
        super(MPEC1_Transformation, self).__init__()

    def _apply_to(self, instance, **kwds):
        options = kwds.pop('options', {})
        bound = kwds.pop('mpec_bound', 0.0)
        #
        # Create a mutable parameter that defines the value of the upper bound
        # on the constraints
        #
        bound = options.get('mpec_bound', bound)
        instance.mpec_bound = Param(mutable=True, initialize=bound)
        #
        # Setup transformation data
        #
        tdata = instance._transformation_data['mpec.simple_nonlinear']
        tdata.compl_cuids = []
        #
        # Iterate over the model finding Complementarity components
        #
        for complementarity in instance.component_objects(
            Complementarity,
            active=True,
            descend_into=(Block, Disjunct),
            sort=SortComponents.deterministic,
        ):
            block = complementarity.parent_block()
            for index in sorted(complementarity.keys()):
                _data = complementarity[index]
                if not _data.active:
                    continue
                _data.to_standard_form()
                #
                _type = getattr(_data.c, "_complementarity_type", 0)
                if _type == 1:
                    #
                    # Constraint expression is bounded below, so we can replace
                    # constraint c with a constraint that ensures that either
                    # constraint c is active or variable v is at its lower bound.
                    #
                    _data.ccon = Constraint(
                        expr=(_data.c.body - _data.c.lower) * _data.v
                        <= instance.mpec_bound
                    )
                    del _data.c._complementarity_type
                elif _type == 3:
                    #
                    # Variable v is bounded above and below.  We can define
                    #
                    _data.ccon_l = Constraint(
                        expr=(_data.v - _data.v.bounds[0]) * _data.c.body
                        <= instance.mpec_bound
                    )
                    _data.ccon_u = Constraint(
                        expr=(_data.v - _data.v.bounds[1]) * _data.c.body
                        <= instance.mpec_bound
                    )
                    del _data.c._complementarity_type
                elif _type == 2:  # pragma:nocover
                    raise ValueError(
                        "to_standard_form does not generate _type 2 expressions"
                    )
            tdata.compl_cuids.append(ComponentUID(complementarity))
            block.reclassify_component_type(complementarity, Block)
